"""
Functions that operate on CSR matrices

Nigel 3 dec 2020
"""
import copy
import numpy as np
from scipy.sparse import csr_matrix

from typing import Union

from banksy_utils.time_utils import timer


@timer
def remove_greater_than(graph: csr_matrix,
                        threshold: float,
                        copy: bool = False,
                        verbose: bool = True):
    """
    Remove values greater than a threshold from a CSR matrix
    """
    if copy:
        graph = graph.copy()

    greater_indices = np.where(graph.data > threshold)[0]

    if verbose:
        print(f"CSR data field:\n{graph.data}\n"
              f"compressed indices of values > threshold:\n{greater_indices}\n")

    # delete the entries in data and index fields
    # -------------------------------------------

    graph.data = np.delete(graph.data, greater_indices)
    graph.indices = np.delete(graph.indices, greater_indices)

    # update the index pointer
    # ------------------------

    hist, _ = np.histogram(greater_indices, bins=graph.indptr)
    cum_hist = np.cumsum(hist)
    graph.indptr[1:] -= cum_hist

    if verbose:
        print(f"\nCumulative histogram:\n{cum_hist}\n"
              f"\n___ New CSR ___\n"
              f"pointers:\n{graph.indptr}\n"
              f"indices:\n{graph.indices}\n"
              f"data:\n{graph.data}\n")

    return graph


def filter_by_rank_and_threshold(graph: csr_matrix,
                                 max_rank: int = 3,
                                 threshold: Union[float, int] = 5,
                                 copy: bool = True,
                                 verbose: bool = True):
    """
    Filter a csr matrix by removing elements in each ROW
    which are either:
        (1) Among the LARGEST elemtns by rank (up to max_rank)
        (2) HIGHER than or equal to a threshold value
    """

    indptr, indices, data = graph.indptr, graph.indices, graph.data

    # initialize mask (default is to keep all entries)
    keep_mask = np.ones_like(data, dtype=bool)
    # need to update the index pointer since some entries will be deleted
    new_indptr = np.zeros_like(indptr)

    current_ptr = 0  # keep track of the modified index pointer for each row

    for row_num in range(graph.shape[0]):

        start_ptr, end_ptr = indptr[row_num], indptr[row_num + 1]
        num_row_elements = end_ptr - start_ptr

        if num_row_elements > 0:  # make sure row is not empty

            if num_row_elements <= max_rank:
                # keep the whole row, since the worst ranked element
                # is still within the max_rank
                current_ptr += num_row_elements
            else:
                row_data = data[start_ptr:end_ptr]

                max_rank_value = row_data[np.argsort(row_data)[-max_rank]]
                row_mask = (row_data >= max_rank_value) | (row_data >= threshold)

                keep_mask[start_ptr:end_ptr] = row_mask

                current_ptr += sum(row_mask)

                # NOTE: we use the max-ranked value here so that if
                # multiple elements have the max-ranked value
                # all of them will be selected.
                # If we just picked the highest ranked elements up to max_rank
                # we will only get one of these elements

                if verbose:
                    print(f"Value of maximum ranked element in "
                          f"{row_data} is {max_rank_value}. "
                          f"keep mask for this row is: {row_mask}")

        new_indptr[row_num + 1] = current_ptr

    if copy:
        new_graph = csr_matrix((data[keep_mask],
                                indices[keep_mask],
                                new_indptr),
                               shape=graph.shape)
    else:
        graph.data = data[keep_mask]
        graph.indices = indices[keep_mask]
        graph.indptr = new_indptr
        new_graph = graph

    return new_graph


@timer
def row_normalize(graph: csr_matrix,
                  copy: bool = False,
                  verbose: bool = True):
    """
    Normalize a compressed sparse row (CSR) matrix by row
    """
    if copy:
        graph = graph.copy()

    data = graph.data

    for start_ptr, end_ptr in zip(graph.indptr[:-1], graph.indptr[1:]):

        row_sum = data[start_ptr:end_ptr].sum()

        if row_sum != 0:
            data[start_ptr:end_ptr] /= row_sum

        if verbose:
            print(f"normalized sum from ptr {start_ptr} to {end_ptr} "
                  f"({end_ptr - start_ptr} entries)",
                  np.sum(graph.data[start_ptr:end_ptr]))

    return graph


def insert_empty_rows(matrix: csr_matrix,
                      num_rows: int,
                      inplace: bool = False,
                      verbose: bool = True,
                      ) -> csr_matrix:
    """
    insert num_rows empty rows  between every row of the given matrix
    """
    assert num_rows >= 1, (
        f"Number of rows to insert was {num_rows}, must be at least 1"
    )

    if inplace:
        padded_matrix = matrix
    else:
        padded_matrix = copy.copy(matrix)

    if verbose: print(f"\noriginal length of indexptr: {padded_matrix.indptr}\n"
                      f"Shape of original matrix: {padded_matrix.shape}\n")

    padded_matrix.indptr = np.repeat(padded_matrix.indptr, num_rows + 1)[num_rows:-num_rows]
    padded_matrix.shape = (len(padded_matrix.indptr), padded_matrix.shape[1])

    if verbose: print(f"modified length of indexptr: {padded_matrix.indptr}\n"
                      f"Shape of modified matrix: {padded_matrix.shape}\n")

    return padded_matrix


def elements_per_row(matrix: csr_matrix):
    """
    find the number of nonzero elements at each row of the CSR matrix
    """
    indptr = matrix.indptr
    return indptr[1:] - indptr[:-1]


def find_nonempty_rows(matrix: csr_matrix,
                       ) -> np.ndarray:
    """
    find indices of non-empty rows of a CSR matrix
    """
    return np.nonzero(elements_per_row(matrix))[0]


def find_empty_rows(matrix: csr_matrix,
                    ) -> np.ndarray:
    """
    find indices of empty rows of a CSR matrix
    """
    return np.nonzero(elements_per_row(matrix) == 0)[0]


@timer
def onehot_labels_toarray(onehot_labels: np.ndarray,
                          verbose: bool = True,
                          ) -> np.ndarray:
    """
    Convert one-hot labels to dense array of labels

     reordered (matmul) sparse_labels -> dense relabelled array
       e.g.
       [ 1 3 2 ] @ [ 0 ... 1 0  -> [ 2 ... 1 3 ]
                     0 ... 0 1
                     1 ... 0 0]

    Onehot labels must have shape (num_labels x num_samples)
    and all nonzero entries should be 1
    """
    print(np.arange(onehot_labels.shape[0]))
    print(onehot_labels.shape)

    dense_array = (np.arange(onehot_labels.shape[0]) @ onehot_labels).astype(np.int32)

    if verbose:
        print(f"Dense array of labels: {dense_array}")

    return dense_array


@timer
def labels_to_onehot(labels: np.ndarray,
                     verbose: bool = True,
                     normalize_by_numlabels: bool = True,
                     ):
    """
    convert an array of labels to a 
    num_labels x num_samples sparse one-hot matrix

    Labels MUST be integers starting from 0,
    but can have gaps in between e.g. [0,1,5,9]
    """

    assert np.issubdtype(labels.dtype, np.integer), (
        f"labels provided are of type {labels.dtype}. "
        f"Should be integer.\n"
    )

    assert np.amin(labels) >= 0, (
        f"Some of the labels have negative values.\n"
        f"All labels must be 0 or positive integers.\n"
    )

    labels_list = np.unique(labels)
    max_label = np.amax(labels_list)
    # Note: onehot matrix will have max_label + 1 rows because 0 is a label
    num_labels = len(labels_list)  # this is the number of used labels
    num_samples = len(labels)
    num_per_label = []

    # initialize the fields for the CSR
    # ---------------------------------

    indptr = np.zeros((max_label + 2,), dtype=np.int32)
    indices = np.zeros((num_samples,), dtype=np.int32)
    data = np.ones_like(indices, dtype=np.float64)

    if verbose:
        print(f"\n--- {num_labels} labels, {num_samples} samples ---\n"
              f"Maximum label is {max_label}"
              f"initalized {indptr.shape} index ptr: {indptr}\n"
              f"initalized {indices.shape} indices: {indices}\n"
              f"initalized {data.shape} data: {data}\n")

    # update index pointer and indices row by row
    # -------------------------------------------

    for label in range(max_label + 1):

        label_indices = np.nonzero(labels == label)[0]
        num_current_label = len(label_indices)
        num_per_label.append(num_current_label)

        previous_ptr = indptr[label]
        current_ptr = previous_ptr + num_current_label
        indptr[label + 1] = current_ptr

        if verbose:
            print(f"indices for label {label}: {label_indices}\n"
                  f"previous pointer: {previous_ptr}, "
                  f"current pointer: {current_ptr}\n")

        if current_ptr >= previous_ptr:
            indices[previous_ptr:current_ptr] = label_indices

        if normalize_by_numlabels:
            data[previous_ptr:current_ptr] /= num_current_label

    onehot_matrix = csr_matrix((data, indices, indptr),
                               shape=(max_label + 1, num_samples))

    # assert len(num_per_label) == (max_label + 1) , (
    #     "error computing numbers of each label"
    # )

    return onehot_matrix, labels_list, num_per_label



"""
Function to compute noise-equivalent singular value 
for selection of number of PCs, as described in (Moffit et. al. 2018)

Also, functions to plot variance contributions and singular values

Nigel 4 mar 2021
"""

import numpy as np
from sklearn.decomposition import PCA

from banksy_utils.time_utils import timer

from typing import Tuple, List, Union

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle


@timer
def noise_equiv_singular_value(data: np.ndarray,
                               num_permutations: int = 50,
                               average_type: str = "mean",
                               verbose: bool = True,
                               ) -> Tuple[float, np.ndarray]:
    """
    get the noise-equivalent maximum singular value of a data matrix
    Each column will be seperately randomly permuted and singular values computed
    """
    assert data.ndim == 2, f"Data has {data.ndim} dimensions, should have 2"

    n_rows, ncols = data.shape

    # set up random number generator for permuting columns
    rng = np.random.default_rng()

    # shuffle columns (n times) and compute highest singular value
    # ------------------------------------------------------------

    all_singular_values = np.zeros((num_permutations,))

    for n in range(num_permutations):

        temp = data.copy()

        # permute each row seperately
        for col in range(ncols):
            temp[:, col] = rng.permutation(temp[:, col])

        all_singular_values[n] = PCA(n_components=1).fit(temp).singular_values_[0]

    if verbose:
        print(f"List of all permuted top "
              f"singular values:\n{all_singular_values}\n")

    # Average the singular values
    # ---------------------------

    if average_type == "mean":
        return np.mean(all_singular_values), all_singular_values

    elif average_type == "median":
        return np.median(all_singular_values), all_singular_values

    else:
        raise ValueError(
            "Average type not recognised. Should be 'mean' or 'median'."
        )


def plot_singular_values(pca: PCA,
                         noise_highest_sv: Union[float, np.ndarray] = None,
                         title: str = None,
                         figsize: Tuple[int, int] = (6, 3),
                         ax=None,
                         ) -> None:
    """
    Plot variance contribution for each component (elbow plot)
    :param pca: PCA object from scikit-learn (Must already be fit to data)
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure

    singular_values = pca.singular_values_
    num_pcs = len(singular_values)
    pcs = np.arange(1, num_pcs + 1)

    ax.plot(pcs, singular_values, color='royalblue', marker='o')

    if noise_highest_sv is not None:

        if isinstance(noise_highest_sv, float):

            ax.hlines(y=noise_highest_sv,
                      xmin=0, xmax=num_pcs + 1,
                      linewidth=1, color='r')

        elif isinstance(noise_highest_sv, np.ndarray):

            mean_sv = np.mean(noise_highest_sv)
            std_sv = np.std(noise_highest_sv)

            ax.hlines(y=mean_sv,
                      xmin=0, xmax=num_pcs + 1,
                      linewidth=std_sv * 2, color='r', alpha=0.2)
            ax.hlines(y=mean_sv,
                      xmin=0, xmax=num_pcs + 1,
                      linewidth=0.2, color='firebrick')

    ax.set_xticks(pcs)
    ax.set_xticklabels(pcs)
    ax.set_xlabel("number of components")
    ax.set_ylabel("singular values")
    if title is not None:
        ax.set_title(title)


def plot_remaining_variance(pca: PCA,
                            title: str = None,
                            figsize: Tuple[int, int] = (6, 3),
                            ax=None,
                            ) -> None:
    """
    Plot variance contribution for each component (elbow plot)
    :param pca: PCA object from scikit-learn (Must already be fit to data)
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure

    remaining_variance = 1 - np.cumsum(pca.explained_variance_ratio_)
    components = np.arange(1, len(remaining_variance) + 1)
    ax.plot(components, remaining_variance,
            color='forestgreen', marker='o')

    ax.set_xticks(components)
    ax.set_xticklabels(components)
    ax.set_xlabel("number of components")
    ax.set_ylabel("remaining variance")
    if title is not None:
        ax.set_title(title)



"""
initialize the graphs neccesary to run BANKSY for a given number of neighbours,
and vidualize graphs/histograms for QC purposes.

TODO: allow for parameter sweep across multiple num_neighbours settings

Yifei Aug 2023
"""

from typing import Tuple

import anndata

from banksy_utils.plot_utils import plot_edge_histograms, plot_weights, plot_theta_graph
from banksy.main import generate_spatial_weights_fixed_nbrs, median_dist_to_nearest_neighbour


def initialize_banksy(adata: anndata.AnnData,
                      coord_keys: Tuple[str],
                      num_neighbours: int = 15,
                      nbr_weight_decay: str = 'scaled_gaussian',
                      max_m: int = 1,
                      plt_edge_hist: bool = True,
                      plt_nbr_weights: bool = True,
                      plt_agf_angles: bool = False,
                      plt_theta: bool = True
                      ) -> dict:
    '''Main Function that initializes the BANKSY Object as a dictionary
    
    Input Args:
        adata (AnnData): AnnData object containing the data matrix

        num_neighbours or k_geom (int) : The number of neighbours in which the edges,
        weights and theta graph are constructed

        nbr_weight_decay (str): Type of neighbourhood decay function, can be 'scaled_gaussian' or 'reciprocal'

        max_m (int): Maximum order of azimuthal gabor filter, we recommend a default of 1

    
    Optional Args:
        plt_edge (bool): Visualize the edge histogram*

        plt_weight (bool): Visualize the weights graph

        plt_agf_weights (bool): Visualize the AGF weights

        plt_theta (bool): Visualize angles around random cell
    '''

    banksy_dict = {}  # Dictionary containing all the information across different parameter settings during runtime

    weights_graphs = {}  # Sub-dictionary containing weighted graph

    # Find median distance to closest neighbours
    nbrs = median_dist_to_nearest_neighbour(adata, key=coord_keys[2])

    for m in range(max_m + 1):

        weights_graph, distance_graph, theta_graph = generate_spatial_weights_fixed_nbrs(
            adata.obsm[coord_keys[2]],
            m=m,
            num_neighbours=num_neighbours,
            decay_type=nbr_weight_decay,
            nbr_object=nbrs,
            verbose=False,
            max_radius=None
        )

        weights_graphs[m] = weights_graph

        if plt_edge_hist:
            print(f'----- Plotting Edge Histograms for m = {m} -----')
            plot_edge_histograms(
                distance_graph=distance_graph,
                weights_graph=weights_graph,
                decay_type=nbr_weight_decay,
                m=m
            )

    banksy_dict[nbr_weight_decay] = {"weights": weights_graphs}

    if plt_nbr_weights:
        # Plot weights graph (Optional)
        print(f'----- Plotting Weights Graph -----')
        plot_weights(
            adata,
            banksy_dict,
            nbr_weight_decay,
            max_m,
            fig_title=f'Decay Type {nbr_weight_decay}',
            coord_keys=coord_keys
        )

    if plt_agf_angles:
        print(f'----- Plotting Azimuthal Angles -----')
        plot_weights(
            adata,
            banksy_dict,
            nbr_weight_decay,
            max_m,
            fig_title=f'Azimuthal Angles',
            coord_keys=coord_keys,
            theta_graph=theta_graph
        )

    if plt_theta:
        # plot angles of neigbhours around a random cell
        print(f'----- Plotting theta Graph -----')
        plot_theta_graph(
            adata,
            theta_graph,
            coord_keys
        )

    return banksy_dict



"""
Main BANKSY functions

Nigel 3 dec 2020

updated 4 mar 2022
updated 8 Sep 2022
"""

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from typing import Union, Tuple, List

import scipy.sparse as sparse
from scipy.sparse import csr_matrix, issparse
from sklearn.neighbors import NearestNeighbors

import anndata

import igraph
# print(f"Using igraph version {igraph.__version__}")
import leidenalg
# print(f"Using leidenalg version {leidenalg.__version__}\n")

from banksy_utils.time_utils import timer
from banksy.csr_operations import remove_greater_than, row_normalize, filter_by_rank_and_threshold
from banksy.labels import Label

def gaussian_weight_1d(distance: float, sigma: float):
    """
    Calculate normalized gaussian value for a given distance from central point
    Normalized by root(2*pi) x sigma
    """
    return np.exp(-0.5 * distance ** 2 / sigma ** 2) / (sigma * np.sqrt(2 * np.pi))

def gaussian_weight_2d(distance: float, sigma: float):
    """
    Calculate normalized gaussian value for a given distance from central point
    Normalized by 2*pi*sigma-squared
    """
    sigma_squared = float(sigma) ** 2
    return np.exp(-0.5 * distance ** 2 / sigma_squared) / (sigma_squared * 2 * np.pi)

def plot_1d_gaussian(sigma: float, min_val: float, max_val: float, interval: float):
    """
    plot a 1d gaussian distribution, check if area sums to 1
    """
    x = np.arange(min_val, max_val, interval)
    y = gaussian_weight_1d(x, sigma)
    print(f"checking area : {np.sum(y * interval)} (should sum to 1)")
    plt.plot(x, y)

def p_equiv_radius(p: float, sigma: float):
    """
    find radius at which you eliminate fraction p
    of a radial gaussian probability distribution
    """
    assert p < 1.0, f"p was {p}, must be less than 1"
    return np.sqrt(-2 * sigma ** 2 * np.log(p))


#
# spatial graph generation
# ========================
#

@timer
def generate_spatial_distance_graph(locations: np.ndarray,
                                    nbr_object: NearestNeighbors = None,
                                    num_neighbours: int = None,
                                    radius: Union[float, int] = None,
                                    ) -> csr_matrix:
    """
    generate a spatial graph with neighbours within a given radius
    """

    num_locations = locations.shape[0]

    if nbr_object is None:
        # set up neighbour object
        nbrs = NearestNeighbors(algorithm='ball_tree').fit(locations)
    else:  # use provided sklearn NN object
        nbrs = nbr_object

    if num_neighbours is None:
        # no limit to number of neighbours
        return nbrs.radius_neighbors_graph(radius=radius,
                                           mode="distance")

    else:
        assert isinstance(num_neighbours, int), (
            f"number of neighbours {num_neighbours} is not an integer"
        )

        graph_out = nbrs.kneighbors_graph(n_neighbors=num_neighbours,
                                          mode="distance")

        if radius is not None:
            assert isinstance(radius, (float, int)), (
                f"Radius {radius} is not an integer or float"
            )

            graph_out = remove_greater_than(graph_out, radius,
                                            copy=False, verbose=False)

        return graph_out

@timer
def theta_from_spatial_graph(locations: np.ndarray,
                             spatial_graph: csr_matrix,
                             ):
    """
    get azimuthal angles from spatial graph and coordinates
    (assumed dim 1: x, dim 2: y, dim 3: z...)

    returns CSR matrix with theta (azimuthal angles) as .data
    """

    theta_data = np.zeros_like(spatial_graph.data, dtype=np.float32)

    for n in range(spatial_graph.indptr.shape[0] - 1):
        ptr_start, ptr_end = spatial_graph.indptr[n], spatial_graph.indptr[n + 1]
        nbr_indices = spatial_graph.indices[ptr_start:ptr_end]

        self_coord = locations[[n], :]
        nbr_coord = locations[nbr_indices, :]
        relative_coord = nbr_coord - self_coord

        theta_data[ptr_start:ptr_end] = np.arctan2(
            relative_coord[:, 1], relative_coord[:, 0])

    theta_graph = spatial_graph.copy()
    theta_graph.data = theta_data

    return theta_graph


@timer
def generate_spatial_weights_fixed_nbrs(locations: np.ndarray,
                                        m: int = 0,  # azimuthal transform order
                                        num_neighbours: int = 10,
                                        decay_type: str = "reciprocal",
                                        nbr_object: NearestNeighbors = None,
                                        verbose: bool = True,
                                        max_radius: int = None,
                                        ) -> Tuple[csr_matrix, csr_matrix, Union[csr_matrix, None]]:
    """
    generate a graph (csr format) where edge weights decay with distance
    """

    distance_graph = generate_spatial_distance_graph(
        locations,
        nbr_object=nbr_object,
        num_neighbours=num_neighbours * (m + 1),
        radius=max_radius,
    )

    if m > 0:
        theta_graph = theta_from_spatial_graph(locations, distance_graph)
    else:
        theta_graph = None

    graph_out = distance_graph.copy()

    # compute weights from nbr distances (r)
    # --------------------------------------

    if decay_type == "uniform":

        graph_out.data = np.ones_like(graph_out.data)

    elif decay_type == "reciprocal":

        graph_out.data = 1 / graph_out.data

    elif decay_type == "reciprocal_squared":

        graph_out.data = 1 / (graph_out.data ** 2)

    elif decay_type == "scaled_gaussian":

        indptr, data = graph_out.indptr, graph_out.data

        for n in range(len(indptr) - 1):

            start_ptr, end_ptr = indptr[n], indptr[n + 1]
            if end_ptr >= start_ptr:
                # row entries correspond to a cell's neighbours
                nbrs = data[start_ptr:end_ptr]
                median_r = np.median(nbrs)
                #### Changed here
                weights = np.exp(-(nbrs / median_r) ** 2)
                data[start_ptr:end_ptr] = weights

    elif decay_type == "ranked":

        linear_weights = np.exp(
            -1 * (np.arange(1, num_neighbours + 1) * 1.5 / num_neighbours) ** 2
        )

        indptr, data = graph_out.indptr, graph_out.data

        for n in range(len(indptr) - 1):

            start_ptr, end_ptr = indptr[n], indptr[n + 1]

            if end_ptr >= start_ptr:
                # row entries correspond to a cell's neighbours
                nbrs = data[start_ptr:end_ptr]

                # assign the weights in ranked order
                weights = np.empty_like(linear_weights)
                weights[np.argsort(nbrs)] = linear_weights

                data[start_ptr:end_ptr] = weights

    else:
        raise ValueError(
            f"Weights decay type <{decay_type}> not recognised.\n"
            f"Should be 'uniform', 'reciprocal', 'reciprocal_squared', "
            f"'scaled_gaussian' or 'ranked'."
        )

    # make nbr weights sum to 1
    graph_out = row_normalize(graph_out, verbose=verbose)

    # multiply by Azimuthal Fourier Transform
    if m > 0:
        graph_out.data = graph_out.data * np.exp(1j * m * theta_graph.data)

    return graph_out, distance_graph, theta_graph


@timer
def generate_spatial_weights_fixed_radius(locations: np.ndarray,
                                          p: float = 0.05,
                                          sigma: float = 100,
                                          decay_type: str = "gaussian",
                                          nbr_object: NearestNeighbors = None,
                                          max_num_neighbours: int = None,
                                          verbose: bool = True,
                                          ) -> Tuple[csr_matrix, csr_matrix]:
    """
    generate a graph (csr format) where edge weights decay with distance
    """

    if decay_type == "gaussian":
        r = p_equiv_radius(p, sigma)
        if verbose:
            print(f"Equivalent radius for removing {p} of "
                  f"gaussian distribution with sigma {sigma} is: {r}\n")
    else:
        raise ValueError(
            f"decay_type {decay_type} incorrect or not implemented")

    distance_graph = generate_spatial_distance_graph(locations,
                                                     nbr_object=nbr_object,
                                                     num_neighbours=max_num_neighbours,
                                                     radius=r)

    graph_out = distance_graph.copy()

    # convert distances to weights
    if decay_type == "gaussian":
        graph_out.data = gaussian_weight_2d(graph_out.data, sigma)
    else:
        raise ValueError(
            f"decay_type {decay_type} incorrect or not implemented")

    return row_normalize(graph_out, verbose=verbose), distance_graph


#
# Combining self / neighbours
# ===========================
#
def zscore(matrix: Union[np.ndarray, csr_matrix],
           axis: int = 0,
           ) -> np.ndarray:
    """
    Z-score data matrix along desired dimension
    """

    E_x = matrix.mean(axis=axis)

    if issparse(matrix):

        squared = matrix.copy()
        squared.data **= 2
        E_x2 = squared.mean(axis=axis)

    else:

        E_x2 = np.square(matrix).mean(axis=axis)

    variance = E_x2 - np.square(E_x)

    zscored_matrix = (matrix - E_x) / np.sqrt(variance)

    if isinstance(zscored_matrix, np.matrix):
        zscored_matrix = np.array(zscored_matrix)

    # Ensure that there are no NaNs
    # (which occur when all values are 0, hence variance is 0)
    zscored_matrix = np.nan_to_num(zscored_matrix)

    return zscored_matrix

def concatenate_all(matrix_list: List[Union[np.ndarray, csr_matrix]],
                    neighbourhood_contribution: float,
                    adata: anndata.AnnData = None,
                    ) -> np.ndarray:
    """
    Concatenate self- with neighbour- feature matrices
    converts all matrices are dense
    z-scores each matrix and applies relevant neighbour contributions (lambda)
    each higher k is given half the weighting of the previous one.
    """

    num_k = len(matrix_list) - 1

    scale_factors_squared = np.zeros(len(matrix_list))

    scale_factors_squared[0] = 1 - neighbourhood_contribution

    denom = 0
    for k in range(num_k):
        denom += 1 / (2 ** (k + 1))

    for k in range(num_k):
        scale_factors_squared[k + 1] = 1 / (2 ** (k + 1)) / denom * neighbourhood_contribution

    scale_factors = np.sqrt(scale_factors_squared)

    print(f"Scale factors squared: {scale_factors_squared}\nScale factors: {scale_factors}")

    scaled_list = []
    for n in range(len(matrix_list)):
        mat = matrix_list[n]
        if issparse(mat):
            mat = mat.todense()
        if np.iscomplexobj(mat):
            mat = np.absolute(mat)
        scaled_list.append(scale_factors[n] * zscore(mat, axis=0))

    concatenated_matrix = np.concatenate(scaled_list, axis=1)

    if isinstance(adata, anndata.AnnData):

        var_original = adata.var.copy()
        var_original["is_nbr"] = False
        var_original["k"] = -1  # k not applicable

        var_list = [var_original, ]
        for k in range(num_k):
            var_nbrs = adata.var.copy()
            var_nbrs["is_nbr"] = True
            var_nbrs["k"] = k
            var_nbrs.index += f"_nbr_{k}"
            var_list.append(var_nbrs)

        var_combined = pd.concat(var_list)

        return anndata.AnnData(concatenated_matrix, obs=adata.obs, var=var_combined)

    elif adata is None:
        return concatenated_matrix

    else:
        print("Adata type not recognised. Should be AnnData or None.")


@timer
def weighted_concatenate(cell_genes: Union[np.ndarray, csr_matrix],
                         neighbours: Union[np.ndarray, csr_matrix],
                         neighbourhood_contribution: float,
                         ) -> Union[np.ndarray, csr_matrix]:
    """
    Concatenate self- with neighbour- feature matrices
    with a given contribution towards disimilarity from the neighbour features (lambda).
    Assumes that both matrices have already been z-scored.
    Will do sparse concatenation if BOTH matrices are sparse.
    """
    cell_genes *= np.sqrt(1 - neighbourhood_contribution)
    neighbours *= np.sqrt(neighbourhood_contribution)

    if issparse(cell_genes) and issparse(neighbours):

        return sparse.hstack((cell_genes, neighbours))

    else:  # at least one is a dense array

        if issparse(cell_genes):
            cell_genes = cell_genes.todense()
        elif issparse(neighbours):
            neighbours = neighbours.todense()

        return np.concatenate((cell_genes, neighbours), axis=1)

def banksy_matrix_to_adata(banksy_matrix,
                           adata: anndata.AnnData,  # original adata object
                           ) -> anndata.AnnData:
    """
    convert a banksy matrix to adata object, by 
     - duplicating the original var (per-gene) annotations and adding "_nbr"
     - keeping the obs (per-cell) annotations the same as original anndata that banksy matrix was computed from
    """

    var_nbrs = adata.var.copy()
    var_nbrs.index += "_nbr"
    nbr_bool = np.zeros((var_nbrs.shape[0] * 2,), dtype=bool)
    nbr_bool[var_nbrs.shape[0]:] = True
    print("num_nbrs:", sum(nbr_bool))

    var_combined = pd.concat([adata.var, var_nbrs])
    var_combined["is_nbr"] = nbr_bool

    return anndata.AnnData(banksy_matrix, obs=adata.obs, var=var_combined)


#
# Leiden Clustering
# ========================
#

class LeidenPartition(object):
    """
    Do leiden partitioning on an input space (can be cell x gene or BANKSY)
    should be reduced by PCA first before feeding into this function
    """

    def __init__(self,
                 input_space: np.ndarray,
                 num_nn: int = 20,
                 nns_have_weights: bool = True,
                 compute_shared_nn: bool = True,
                 filter_shared_nn: bool = True,
                 shared_nn_max_rank: int = 3,
                 shared_nn_min_shared_nbrs: int = 5,
                 nn_algorithm: str = "ball_tree",
                 verbose: bool = True,
                 ) -> None:
        """
        num_nn: number of nearest neighbours
        nns_have_weights: whether to weight the nearest neighbour edges
                                 edges will be 1 / (1 + distance)
        compute_shared_nn: whether to compute shared nearest neighbours
        nn_algorithm: same as sk-learn's NN algorithm input
        """

        self.verbose = verbose
        self.nn_algorithm = nn_algorithm

        self.num_nn = num_nn
        self.input_space = input_space

        # compute nearest neighbour graph (and weights)
        # ---------------------------------------------

        self.nns_have_weights = nns_have_weights

        if self.nns_have_weights:

            self.nbrs, self.nn_weighted = self.find_nn(
                weighted=True,
            )

            # convert weights to connectivity (ones entries)
            self.nn_connectivity = self.nn_weighted.copy().astype(np.int16)
            self.nn_connectivity.data = np.ones_like(
                self.nn_connectivity.data,
            )
            self._print_csr_info(self.nn_connectivity,
                                 "Nearest-neighbour connectivity")

        else:
            self.nbrs, self.nn_connectivity = self.find_nn(
                weighted=False,
            )
            self.nn_weighted = None

        # Compute Shared Nearest Neighbours
        # ---------------------------------

        if compute_shared_nn:

            self.shared_nn_max_rank = shared_nn_max_rank
            self.shared_nn_min_shared_nbrs = shared_nn_min_shared_nbrs

            self.snn_connections, self.snn_graph = self.shared_nn(
                self.nn_connectivity,
                allow_only_nn=True,
                filter=filter_shared_nn,
                max_rank=self.shared_nn_max_rank,
                threshold=self.shared_nn_min_shared_nbrs,
            )

            if self.nns_have_weights:

                print(f"\n-- Multiplying sNN connectivity by weights --\n")

                self.snn_weighted = self.nn_weighted.multiply(
                    self.snn_connections  # .astype(np.float64)
                )
                self._print_csr_info(self.snn_weighted,
                                     "shared NN with distance-based weights")
                print(
                    f"shared NN weighted graph data: {self.snn_weighted.data}")
                self.G = self.csr_to_igraph(self.snn_weighted, )

            else:
                self.snn_weighted = None
                # self.snn_connections = self.snn_connections.multiply(
                #     self.nn_connectivity
                # )
                self.G = self.csr_to_igraph(self.snn_connections, )

        else:

            self.snn_connections, self.snn_graph = None, None
            self.snn_weighted = None

            if self.nns_have_weights:
                self.G = self.csr_to_igraph(self.nn_weighted, )
            else:
                self.G = self.csr_to_igraph(self.nn_connectivity, )

    def _print_csr_info(self,
                        csr: csr_matrix,
                        name: str,
                        ):
        print(f"\n{name} graph "
              f"(dtype: {csr.dtype}, shape: {csr.shape}) "
              f"has {csr.nnz} nonzero entries.")

    @timer
    def find_nn(self,
                weighted: bool = True,
                ) -> Tuple[csr_matrix, csr_matrix]:

        nbrs = NearestNeighbors(
            algorithm=self.nn_algorithm,
        ).fit(self.input_space)

        if weighted:
            mode = "distance"
        else:
            mode = "connectivity"

        nn_graph = nbrs.kneighbors_graph(
            n_neighbors=self.num_nn,
            mode=mode,
        )

        if weighted:
            # compute weights from distances
            nn_graph.data = 1 / (nn_graph.data + 1)

        if self.verbose:
            if weighted:
                graph_name = "Nearest-neighbour weighted"
            else:
                graph_name = "Nearest-neighbour unweighted"
            self._print_csr_info(nn_graph, graph_name)

        return nbrs, nn_graph

    @timer

    def shared_nn(self,
                  nn_connectivity: csr_matrix,
                  filter: bool = True,
                  allow_only_nn: bool = True,
                  max_rank: int = 3,
                  threshold: Union[float, int] = 5,
                  ) -> Tuple[csr_matrix, csr_matrix]:
        """
        compute shared nearest neighbour graph 
        from nearest neighbour graph
        """
        # compute all shared nearest neighbours
        snn_graph = nn_connectivity @ nn_connectivity.T

        if allow_only_nn:
            original_nnz = snn_graph.nnz

            snn_graph = snn_graph.multiply(nn_connectivity)

            print(f"\n(after computing shared NN)\n"
                  f"Allowing nearest neighbours only reduced the number of "
                  f"shared NN from {original_nnz} to {snn_graph.nnz}.\n")

        if filter:
            # filter out
            # shared nns with low rank (i.e. fewest shared neighbours)
            # or below threshold number of shared neighbours
            snn_graph = filter_by_rank_and_threshold(
                snn_graph,
                max_rank=max_rank,
                threshold=threshold,
                copy=False,
                verbose=False,
            )

        # convert snn_filtered to binary (i.e. connectivity only)
        snn_connections = snn_graph.copy()
        snn_connections.data = np.ones_like(snn_connections.data)

        if self.verbose:
            self._print_csr_info(
                snn_connections,
                "Shared nearest-neighbour (connections only)"
            )
            self._print_csr_info(
                snn_graph,
                "Shared nearest-neighbour (number of shared neighbours as weights)"
            )
            print(f"\nsNN graph data:\n{snn_graph.data}\n")

        return snn_connections, snn_graph

    @timer

    def csr_to_igraph(self,
                      csr_mat: csr_matrix,
                      ) -> igraph.Graph:
        """
        convert a CSR matrix (scipy) into native iGraph format
        """

        self._print_csr_info(csr_mat, "Converting")

        sources, targets = csr_mat.nonzero()
        edgelist = zip(sources.tolist(), targets.tolist())

        return igraph.Graph(edgelist,
                            edge_attrs={"weight": csr_mat.data.tolist()})

    @timer

    def partition(self,
                  resolution: float = 0.7,
                  partition_metric=leidenalg.RBConfigurationVertexPartition,
                  n_iterations: int = 100,
                  seed: int = None,
                  ) -> Tuple[Label, float]:
        """
        partition the graph
        """
        partition = leidenalg.find_partition(
            self.G,
            partition_metric,
            resolution_parameter=resolution,
            weights="weight",
            n_iterations=n_iterations,
            seed=seed,
        )

        label = Label(partition.membership)

        print(f"---- Partitioned BANKSY graph ----\n"
              f"modularity: {partition.modularity :0.2f}\n"
              f"{label.num_labels} unique labels:\n{label.ids}\n")

        return label, partition.modularity

@timer
def median_dist_to_nearest_neighbour(adata: anndata.AnnData,
                                     key: str = "coord_xy"):
    
    '''Finds and returns median cell distance in a graph'''
    nbrs = NearestNeighbors(algorithm='ball_tree').fit(adata.obsm[key])
    distances, indices = nbrs.kneighbors(n_neighbors=1)
    median_cell_distance = np.median(distances)
    print(f"\nMedian distance to closest cell = {median_cell_distance}\n")
    return nbrs


"""
Create BANKSY Matrix from annadata

Yifei May 2023
"""
import os, time, gc
import numpy as np
import matplotlib.pyplot as plt

import scipy as sp
import scipy.sparse as sparse
from scipy.sparse import csr_matrix, issparse

import anndata
from copy import deepcopy
from datetime import datetime
from banksy.main import concatenate_all
from sklearn.decomposition import PCA
from typing import Tuple


def generate_banksy_matrix(adata: anndata.AnnData,
                           banksy_dict: dict,
                           lambda_list: list,
                           max_m: int,
                           plot_std: bool = False,
                           save_matrix: bool = False,
                           save_folder: str = './data',
                           variance_balance: bool = False,
                           verbose: bool = True) -> Tuple[dict, np.ndarray]:
    '''Creates the banksy matrices with the set hyperparameters given.
    Stores the computed banksy matrices in the banksy_dict object
    Returns the *last* banksy matrix that was computed.'''

    # Print time of ran
    time_str = datetime.now().strftime("%b-%d-%Y-%H-%M")
    # Print gene_list
    gene_list = adata.var.index
    if verbose:
        print(f'Runtime {time_str}')
        print(f"\n{len(gene_list)} genes to be analysed:\nGene List:\n{gene_list}\n")

    for nbr_weight_decay in banksy_dict:

        # First create neighbour matrices
        nbr_matrices = create_nbr_matrix(adata, banksy_dict, nbr_weight_decay, max_m, variance_balance)

        # Create matrix list
        mat_list, concatenated = create_mat_list(adata, nbr_matrices, max_m)

        # save un-nomalized (i.e. no lambda factor) version of BANKSY matrix
        banksy_dict[nbr_weight_decay]["norm_counts_concatenated"] = concatenated

        if verbose:
            print(f"\nCell by gene matrix has shape {adata.shape}\n")

        for lambda_param in lambda_list:
            gc.collect()
            # Create BANKSY matrix by concatenating all 
            banksy_matrix = concatenate_all(mat_list, lambda_param, adata)

            if verbose:
                print(f"Shape of BANKSY matrix: {banksy_matrix.shape}\nType of banksy_matrix: {type(banksy_matrix)}\n")

            # plot standard deviations per gene / nbr gene
            if plot_std:
                st_dev_pergene = convert2dense(banksy_matrix.X).std(axis=0)
                plot_std_per_gene(st_dev_pergene, lambda_param)

            # save as a new AnnData object
            # ----------------------------

            banksy_dict[nbr_weight_decay][lambda_param] = {"adata": banksy_matrix, }

            # save the banksy matrix as a csv if needed
            # -------------------------------
            if save_matrix:
                try:
                    file_name = f"adata_{nbr_weight_decay}_l{lambda_param}_{time_str}.csv"

                    if not os.path.exists(save_folder):
                        print(f"Making save-folder at {save_folder}")
                        os.makedirs(save_folder)

                    banksy_matrix.write(filename=file_name)
                    print(f'Wrote Banksy_file: {file_name} at {save_folder}')

                except PermissionError:
                    print("\nWARNING: Permission denied to save file. Unable to save adata.\n")

    return banksy_dict, banksy_matrix


def create_nbr_matrix(adata,
                      banksy_dict: dict,
                      nbr_weight_decay: str,
                      max_m: int,
                      center: bool = True,
                      variance_balance: bool = False,
                      verbose: bool = True):
    '''Computes the neighbour averaged feature matrices'''
    # Create neighbour matrices
    start_time = time.perf_counter()

    n_components = 20
    X_dense = convert2dense(adata.X)

    if verbose:
        # Show attributes of current cell-gene matrix
        print(f"Decay Type: {nbr_weight_decay}")
        print(f"Weights Object: {banksy_dict[nbr_weight_decay]}")
        # print(f'\nMean and Std of Cells matrix {round(X_dense.mean(),2)} | {round(X_dense.std(),2)}')

    if variance_balance:
        ## We need to balance the variance of the cell's expression
        ## With the variance of the neighbour expression matrix
        pca_own = PCA(n_components=n_components)
        # Zscore across the whole row (axis = 0)

        pca_own.fit_transform(np.nan_to_num(sp.stats.zscore(X_dense, axis=0), nan=0))
        sum_of_cell_pca_var = np.sum(np.square(pca_own.singular_values_))

        if verbose:
            print(f'\nSize of Own | Shape: {X_dense.shape}\nTop 3 entries of Cell Mat:')
            print(X_dense[:3, :3])
            print(f'\nVariance of top {n_components} PCs of cells\'own expression: {round(sum_of_cell_pca_var, 2)}')

    # m == 0, the Nbr matrix, before AGF
    nbr_matrices = {}
    nbr_mat_0 = banksy_dict[nbr_weight_decay]['weights'][0] @ X_dense
    nbr_matrices[0] = nbr_mat_0

    if verbose:
        print(f'\nNbr matrix | Mean: {round(nbr_mat_0.mean(), 2)} | Std: {round(nbr_mat_0.std(), 2)}')
        print(f'Size of Nbr | Shape: {nbr_mat_0.shape}\nTop 3 entries of Nbr Mat:\n')
        print(nbr_mat_0[:3, :3])

    # for m >= 1, AGF is implemented
    for m in range(1, max_m + 1):

        weights = banksy_dict[nbr_weight_decay]["weights"][m]
        weights = weights.copy()

        #weights_abs.data = np.absolute(weights_abs.data)
        weights_data = weights.data
        #nbr_avgs = weights_abs @ X_dense
        nbr_mat = np.zeros(adata.X.shape, )

        # for n in range(weights.indptr.shape[0] - 1):
        #     ind_temp = weights.indices[weights.indptr[n]:weights.indptr[n + 1]]
        #     weight_temp = weights.data[weights.indptr[n]:weights.indptr[n + 1]]
        #     zerod = X_dense[ind_temp, :] - nbr_avgs[n, :]
        #     nbr_mat[n, :] = np.absolute(np.expand_dims(weight_temp, axis=0) @ zerod)
        
        ### Edit: Update Sep 2024 to match Banksy R


        ## use function to avoid -> UnboundLocalError: local variable 'fscale_exp' referenced before assignment
        def fscale_exp (X_dense, ind_temp):
            fscaled_exp = X_dense[ind_temp,] - np.mean(X_dense[ind_temp,],
                                                      axis=0)  # raw expression - mean of expression of each cell in the n-cell neighborhood
            return fscaled_exp


        for n in range(weights.indptr.shape[0] - 1):
            #weights indices corresponds to cell index
            ind_temp = weights.indices[weights.indptr[n]:weights.indptr[n + 1]]

            if center:
                nbr_mat[n, :] = np.absolute(np.dot(weights_data[weights.indptr[n]:weights.indptr[n + 1]], fscale_exp(X_dense, ind_temp)))  # cell expression now contain AGF weights
            else:
                nbr_mat[n, :] = np.absolute(np.dot(weights_data[weights.indptr[n]:weights.indptr[n + 1]], X_dense[ind_temp,]))

        nbr_matrices[m] = nbr_mat
        gc.collect()

        if verbose:
            print(f'\nAGF matrix | Mean: {round(nbr_mat.mean(), 2)} | Std: {round(nbr_mat.std(), 2)}')
            print(f'Size of AGF mat (m = {m}) | Shape: {nbr_mat.shape}')
            print(f'Top entries of AGF:\n{nbr_mat[:3, :3]}')

    if variance_balance:

        for m in range(0, max_m + 1):
            # Balance variance for (neighbourhood / AGF) matrix
            pca_nbr = PCA(n_components=n_components)
            # Zscore across the whole row
            nbr_mat = deepcopy(nbr_matrices[m])
            print(f'\nBalancing (Nbr / AGF) matrix (m = {m}), Variance of Nbr_mat = {round(nbr_mat.var(), 3)}')
            pca_nbr.fit_transform(np.nan_to_num(sp.stats.zscore(nbr_mat, axis=0), nan=0))

            sum_of_nbr_pca_var = np.sum(np.square(pca_nbr.singular_values_))
            balance_factor = np.sqrt(sum_of_cell_pca_var / sum_of_nbr_pca_var)
            nbr_matrices[m] = balance_factor * nbr_mat

            if verbose:
                print(f'Performed Variance Balancing')
                print(f'Variance of top {n_components} PCs of nbrs\' expression {round(sum_of_nbr_pca_var, 2)}')
                print(f'Multiplied nbrs\' expression by balance factor: {round(balance_factor, 2)}')
                print(f'After, Variance of Nbr_mat = {round(nbr_matrices[m].var(), 2)}')

    gc.collect()
    print(f'Ran \'Create BANKSY Matrix\' in {round((time.perf_counter() - start_time) / 60, 2)} mins')
    return nbr_matrices


def create_mat_list(
        adata,
        nbr_matrices,
        max_m: int
):
    '''Auxiliary function to create a list of neighbouring matrices.
    Combines original expression matrix with neighbour matrices (mean and AGF)
    into a list of matrices.
    Also concatenates the matrices into an (unscaled) single matrix.'''

    mat_list = [adata.X]
    for m in range(max_m + 1):
        mat_list.append(nbr_matrices[m])

    # Also save the concatenated un-weighted data for visualization later
    if issparse(adata.X):
        concatenated = sparse.hstack(mat_list, )
    else:
        concatenated = np.concatenate(mat_list, axis=1, )

    return mat_list, concatenated


def plot_std_per_gene(st_dev_pergene, lambda_param, **kwargs):
    '''Plots the standard deviation per gene'''
    options = {'figsize': (8, 2),
               'width': 1,
               'color': 'slateblue'}
    options.update(kwargs)

    fig_title = f"Standard deviations for neighbourhood contribution = {lambda_param}"
    fig, ax = plt.subplots(figsize=options['figsize'])

    ax.bar(np.arange(len(st_dev_pergene)), st_dev_pergene,
           width=options['width'],
           color=options['color'],
           linewidth=0)

    ax.set_title(fig_title)


def convert2dense(X):
    '''Auxiliary function to convert sparse matrix to dense'''
    if isinstance(X, np.ndarray):
        return X
    if issparse(X) or isinstance(X, anndata._core.views.ArrayView):
        X = X.toarray()
    else:
        raise TypeError("sparse format not recognised")

    print('Check if X contains only finite (non-NAN) values')
    assert np.isfinite(X).all()

    return X



"""
Decorator to measure and display execution time of a function

Nigel 3 dec 2020
"""

import time

def timer(func):
    
    def _timer(*args, **kwargs):
        
        start_time = time.time()
        
        try:
            return func(*args, **kwargs)
        
        finally:
            time_taken = time.time() - start_time
            print(f"---- Ran {func.__name__} "
                  f"in {time_taken:0.2f} s ----\n")

        return result
    
    return _timer



"""
This script contains the function to carry out dimensionality reduction on the banksy matrices via PCA and UMAP

Yifei May 2023
"""
from sklearn.decomposition import PCA
import umap
import anndata
import scipy.sparse as sparse
from scipy.sparse import csr_matrix, issparse
from banksy_utils.pca import plot_remaining_variance
from typing import List
import numpy as np

def pca_umap(banksy_dict: dict,
                  pca_dims: List[int] = [20,],
                  plt_remaining_var: bool = True,
                  add_umap: bool = False,
                  **kwargs) -> None:
    '''
    PCA_UMAP first applies dimensionality reduction via PCA,
    then applies UMAP to cluster the groups

    Args:
        banksy_dict (dict): The processing dictionary containing info about the banksy matrices
    
    Optional Arg:
        pca_dims (List of integers): A list of integers which the PCA will reduce to
    
    Variable Args (kwargs):
        figsize (tuple of integers): A tuple for adjusting figure size

    Returns: Plot of remaining variance 
    '''
    options = {
        'figsize': (8,3)
    }
    print(f'Current decay types: {list(banksy_dict.keys())}')

    for nbr_weight_decay in banksy_dict:
        
        for lambda_param in banksy_dict[nbr_weight_decay]:
            
            if isinstance(lambda_param, str):
                continue # skip weights matrices

            print(
                f"\nReducing dims of dataset in (Index = {nbr_weight_decay}, lambda = {lambda_param})\n"
                + "=" * 50 + "\n"
            )
            
            # Retrieve anndata object
            # -----------------------
            
            adata_temp = banksy_dict[nbr_weight_decay][lambda_param]["adata"]
            
            X = adata_temp.X.todense() if issparse(adata_temp.X) else adata_temp.X
            
            X[np.isnan(X)] = 0

            # Reduce dimensions by PCA and then UMAP
            # --------------------------------------
                
            for pca_dim in pca_dims:

                if isinstance(pca_dim, int):
                    # We manually specify the number of PCs
                    print(f"Setting the total number of PC = {pca_dim}")
                    pca = PCA(n_components=pca_dim)
                    reduced = pca.fit_transform(X)
                
                elif isinstance(pca_dim, float):
                    # Otherwise, we specify the number of cumulative variance that each PC should represent
                    print(f"Setting the total cumulative variance of PCs = {pca_dim}")
                    pca = PCA(n_components=pca_dim)
                    reduced = pca.fit_transform(X)

                #print(f'Noise Variance of PCA: {pca.noise_variance_}')
                #print(f'Variance of each PCs: {pca.explained_variance_ratio_}')
                print(
                    f"Original shape of matrix: {X.shape}"
                    f"\nReduced shape of matrix: {reduced.shape}\n" + "-" * 60 
                    + f"\nmin_value = {reduced.min()}, mean = {reduced.mean()}, max = {reduced.max()}\n"
                )

                adata_temp.obsm[f"reduced_pc_{pca_dim}"] = reduced
                
                # Plot variance contribution for each component (elbow plot)
                # ----------------------------------------------------------
                if plt_remaining_var:
                    plot_remaining_variance(
                        pca, 
                        figsize=options["figsize"], 
                        title=f"decay type = {nbr_weight_decay}, lambda = {lambda_param}"
                    )

                # UMAP
                # ----

                if add_umap:
                    print(f'Conducting UMAP and adding embeddings to adata.obsm["reduced_pc_{pca_dim}_umap"]')
                    reducer = umap.UMAP(transform_seed = 42)
                    umap_embedding = reducer.fit_transform(reduced)
                    print(f"UMAP embedding\n" + "-" * 60  + f"\nshape: {umap_embedding.shape}\n\n")

                    # save in AnnData object[obsm] attribute
                    # ------------------
                    adata_temp.obsm[f"reduced_pc_{pca_dim}_umap"] = umap_embedding                    
                    print(adata_temp.obsm)

    


"""
High-level function for generating all relevant plots for each BANKSY/non-spatial result

refactored from ipython notebook in earlier version of code.

-Yifei/Nigel 21 Aug 23
"""

import os, gc
import matplotlib.pyplot as plt
import scanpy as sc

from typing import Union, List, Tuple
import numpy as np
import pandas as pd
from banksy_utils.plotting import plot_2d_embeddings, plot_labels_seperately
from banksy.labels import plot_connections, Label
import anndata

import matplotlib.ticker as ticker
from scipy.sparse import csc_matrix, csr_matrix


def plot_results(
        results_df: pd.DataFrame,
        weights_graph: Union[csc_matrix, csr_matrix],
        c_map: str,
        match_labels: bool,
        coord_keys: Tuple[str],
        max_num_labels: int = 20,
        save_fig: bool = False,
        save_fullfig: bool = False,
        save_seperate_fig: bool = False,
        save_path: str = None,
        plot_dot_plot: bool = False,
        plot_heat_map: bool = False,
        n_genes: int = 5,  # number of genes to use for heatmap/dotplot
        color_list: List[str] = [],
        dataset_name: str = "slideseq_mousecerebellum",
        main_figsize: Tuple[float, float] = (15, 9),
        **kwargs
) -> None:
    '''
    Plot and visualize the results of BankSY 

    Args:
        results_df (pd.DataFrame): DataFrame containing all the results

        weight_graph: weight_graph object in a dictionary

        max_num_labels (int): Maximum number of labels

        match_labels (bool): If the match labels options was previously indicated

        file_path: str, file_path to save plot 

    Optional args (kwargs):

        save_all_h5ad: to save a copy of the temporary anndata object as .h5ad format

        cmap_name : Color map settings for plotting banksy

        file_path (str): default file path is 'data/slide_seq/v1'
    
    Returns:
        The main figure for visualization using banksy
    '''

    options = {
        'save_all_h5ad': False,
        'group_num': 20
    }
    options.update(kwargs)

    for params_name in results_df.index:

        label_index = 'relabeled' if match_labels else 'labels'

        labels = results_df.loc[params_name, label_index]
        adata_temp = results_df.loc[params_name, "adata"]
        num_pcs = results_df.loc[params_name, 'num_pcs']

        pc_temp = adata_temp.obsm[f"reduced_pc_{num_pcs}"]
        umap_temp = adata_temp.obsm[f"reduced_pc_{num_pcs}_umap"]

        label_name = f"labels_{params_name}"
        adata_temp.obs[label_name] = np.char.mod('%d', labels.dense)
        adata_temp.obs[label_name] = adata_temp.obs[label_name].astype('category')

        if options['save_all_h5ad']:
            # If we want to save the anndata object as .h5ad format
            save_path = os.path.join(save_path, f"{dataset_name}_{params_name}.h5ad")
            adata_temp.write(save_path)

        adata_temp.obsm[coord_keys[2]] = np.vstack(
            (adata_temp.obs[coord_keys[0]].values,
             adata_temp.obs[coord_keys[1]].values)
        ).T

        # Plot Main figure
        # ----------------

        fig, grid = _initialize_main_figure(main_figsize=main_figsize)

        # Auxiliary funtion to plot labels
        _plot_labels(adata_temp,
                     key=coord_keys[2],
                     labels=labels,
                     cmap=c_map,
                     color_list=color_list,
                     max_num_labels=max_num_labels,
                     params_name=params_name,
                     fig=fig,
                     grid=grid)

        if save_fig:
            if not os.path.exists(save_path):
                os.makedirs(save_path)

            print(f"Saving figure at {os.path.join(save_path, f'{dataset_name}_{params_name}_spatialmap.png')}")
            fig.savefig(os.path.join(save_path, f'{dataset_name}_{params_name}_spatialmap.png'))

        # Seperate Location plots
        # -----------------------

        plot_labels_seperately(
            labels, adata_temp.obsm[coord_keys[2]],
            embeddings=umap_temp,
            cmap_name=c_map,
            colour_list=color_list,
            default_colour="tab:red",
            plots_per_row=3,
            spot_size=1,
            subplot_size=(5, 5),
            flip_axes=False,
            verbose=False,
            save_fig=save_seperate_fig,
            save_path=os.path.join(save_path, f'{dataset_name}_{params_name}_clusters_seperate.png'),
        )

        # Plot UMAP (again but with labels)
        # ---------------------------------

        ax_umap = fig.add_subplot(grid[0, -2:])
        plot_2d_embeddings(umap_temp, labels.dense,
                           method_str="UMAP",
                           space_str="",
                           xlabel="UMAP 1", ylabel="UMAP 2",
                           ax=ax_umap,
                           cmap_name=c_map,
                           plot_cmap=False,
                           )

        # Plot 1st 2 dimensions of PCA
        # ----------------------------

        dim_sets = (pc_temp[:, :2], pc_temp[:, 1:3])
        dims1 = (0, 1,)
        dims2 = (1, 2,)
        axes = [fig.add_subplot(grid[1, 2 + axnum]) for axnum in range(2)]

        for dim_set, dim1, dim2, ax in zip(dim_sets, dims1, dims2, axes):
            plot_2d_embeddings(dim_set, labels.dense,
                               method_str=f"PCA {dim1 + 1} / {dim2 + 1}",
                               space_str="",
                               xlabel=f"PCA {dim1 + 1}", ylabel=f"PCA {dim2 + 1}",
                               ax=ax,
                               cmap_name=c_map,
                               plot_cmap=False,
                               title_fontsize=9)

        # Plot connectivity between labels
        # --------------------------------

        ax_connections = fig.add_subplot(grid[-1, -2:])

        plot_connections(
            labels,
            weights_graph,
            ax_connections,
            zero_self_connections=True,
            title_str="Connections between label",
            colormap_name=c_map,
        )

        # (optional) Use scanpy functions to plot heatmaps or dotplots
        # ------------------------------------------------------------

        if plot_dot_plot or plot_heat_map:
            num_groups = options['group_num']
            groups_subset = [
                str(n) for n in range(num_groups) if str(n) in adata_temp.obs[label_name].cat.categories
            ]
            print(f"plotting groups: {groups_subset}")

        if plot_heat_map:
            sc.pl.rank_genes_groups_heatmap(adata_temp, n_genes=n_genes,
                                            groups=groups_subset,
                                            vmin=-3, vmax=3, cmap='bwr',
                                            swap_axes=True)

        if plot_dot_plot:
            sc.pl.rank_genes_groups_dotplot(adata_temp, n_genes=n_genes,
                                            groups=groups_subset, )

        if save_fullfig:
            if not os.path.exists(save_path):
                os.makedirs(save_path)

            print(f"Saving full-figure at {os.path.join(save_path, f'{dataset_name}_{params_name}_full_figure.png')}")
            fig.savefig(os.path.join(save_path, f'{dataset_name}_{params_name}_full_figure.png'))


def _initialize_main_figure(main_figsize=(15, 9),
                            width_ratios=(2, 0.1, 0.5, 0.5),
                            height_ratios=(1, 0.3, 1),
                            ) -> Tuple[plt.Figure, plt.grid]:
    '''Auxiliary Function to initialize main figure and associated grids'''

    fig = plt.figure(figsize=main_figsize, constrained_layout=True)

    grid = fig.add_gridspec(ncols=4, nrows=3,
                            width_ratios=width_ratios,
                            height_ratios=height_ratios)

    return fig, grid


def _plot_labels(adata_temp: anndata.AnnData,
                 key: str,
                 labels: Label,
                 cmap: str,
                 color_list: List[str],
                 max_num_labels: int,
                 params_name: str,
                 fig: plt.Figure,
                 grid: plt.grid):
    '''
    Plots the spatial map of cluster labels (each with different colour) in the main figure
    '''

    ax_locs = fig.add_subplot(grid[:, 0])

    # We assign color of points to color_list if they are specified
    if color_list:
        c = [color_list[lbl] for lbl in labels.dense]
    else:
        c = labels.dense

    scatterplot = ax_locs.scatter(adata_temp.obsm[key][:, 0],
                                  adata_temp.obsm[key][:, 1],
                                  c=c,
                                  cmap=cmap,
                                  vmin=0, vmax=max_num_labels - 1,
                                  s=3, alpha=1.0)

    ax_locs.set_aspect('equal', 'datalim')
    ax_locs.set_title(f'BANKSY Labels ({params_name})', fontsize=20, fontweight="bold", )
    # ax_locs.set_ylim(ax_locs.get_ylim()[::-1])

    ax_cbar = fig.add_subplot(grid[:, 1])
    cbar = fig.colorbar(
        scatterplot,
        boundaries=np.arange(max_num_labels + 1) - 0.5,
        cax=ax_cbar,
    )
    cbar.set_ticks(labels.ids)
    cbar.set_ticklabels(labels.ids)
    # Turn of ticks and frame
    ax_locs.set(frame_on=False)
    ax_locs.xaxis.set_major_locator(ticker.NullLocator())
    ax_locs.yaxis.set_major_locator(ticker.NullLocator())



import os, gc
import numpy as np
import anndata
from banksy_utils.slideseq_ref_data import markergenes_dict
import scanpy as sc
import pandas as pd
from sklearn.metrics.cluster import adjusted_rand_score

def pad_clusters(clust2annotation: dict, 
                 original_clusters: list, 
                 pad_name: str = "other"):
    """
    Annotate clusters (in form of a dict) by assigning them annotations
    using clust2annotation dict. Modifies clust2annotation dict in place.

    converts to str if given integers as original clusters.
    
    For any clusters not defined in clust2annotation:
    - assign all with pad_name if str input given
    - keep original name if pad_name is None
    """
    for cluster in original_clusters:
        cluster = str(cluster)
        if cluster not in clust2annotation:
            if pad_name is not None:
                clust2annotation[cluster] = pad_name
            else:
                clust2annotation[cluster] = cluster

def refine_cell_types(adata_spatial,
                      adata_nonspatial,
                      cluster2annotation_refine: dict):
    '''
    Applies the pad_clusters and return the refined clusters 

    Returns refined adata_spatial, adata_nonspatial
    '''
    for adata in [adata_nonspatial, adata_spatial]:

        pad_clusters(cluster2annotation_refine, adata.obs["cell type"], pad_name=None)
        #print(f"New cluster dictionary: {cluster2annotation_refine}")

        adata.obs["cell type refined"] = adata.obs["cell type"].map(
            cluster2annotation_refine
        ).astype('category')
        
    # remove other category
    adata_spatial_filtered = adata_spatial[adata_spatial.obs["cell type refined"] != "other",:]
    adata_nonspatial_filtered = adata_nonspatial[adata_nonspatial.obs["cell type refined"] != "other",:]

    return adata_spatial_filtered, adata_nonspatial_filtered

def create_metagene_df(adata_allgenes: anndata.AnnData,
                       coord_keys: list,
                       markergenes_dict: dict = markergenes_dict):
    '''
    Note that this function works for Slideseq dataset by default
    creates a metagene dataframe from the markergenes dataset from dropviz.org

    can enter your own markergenes dictionary as an optional argument markergenes_dict
    in the form {"cell type": [list of genes...],...}
    '''
    print("Generating Metagene data...\n")

    keys = [coord_keys[0], coord_keys[1]]
    metagene_df = adata_allgenes.obs.loc[:,keys].copy()

    for layer in markergenes_dict:

        markers = [marker for marker in markergenes_dict[layer] 
                if marker in adata_allgenes.var.index]

        print(f"{len(markers)} DE markers for {layer}")

        if markers < markergenes_dict[layer]:
            print(f"{len(markergenes_dict[layer])-len(markers)} "
                f"scRNA-seq DE genes in {layer} absent/filtered from slideseq dataset")

        layer_slice = adata_allgenes[:, markers]

        metagene = np.array(np.mean(layer_slice.X, axis = 1))
        
        metagene_df[layer] = metagene

    return metagene_df

def get_DEgenes(adata, cell_type, top_n=20, verbose=True):
    DE_genes = sc.get.rank_genes_groups_df(adata, group=cell_type)["names"][:top_n].tolist()
    if verbose: print(f"Top {top_n} DE genes for {cell_type} : {DE_genes}\n")
    return DE_genes

def get_metagene_difference(adata, DE_genes1, DE_genes2, m=0):
    """
    Compute the differences in metagene and neighbour metagene
    espression for a particular cell type
    must have run scanpy's rank_genes_groups first
    """
    
    def get_metagene(DE_genes):

        genes_slice = adata[:, DE_genes]
        genes_slice_nbr = adata[:, [gene + f"_nbr_{m:1d}" for gene in DE_genes]]

        metagene = np.mean(genes_slice.X, axis = 1)
        metagene_nbr = np.mean(genes_slice_nbr.X, axis = 1)
        
        return metagene, metagene_nbr
    
    metagene1, metagene_nbr1 = get_metagene(DE_genes1)
    metagene2, metagene_nbr2 = get_metagene(DE_genes2)
    
    return metagene1-metagene2, metagene_nbr1-metagene_nbr2


def create_spatial_nonspatial_adata(results_df: pd.DataFrame,
                                    pca_dims,
                                    lambda_list, 
                                    resolutions,
                                    cluster2annotation_spatial,
                                    cluster2annotation_nonspatial):
    """
    Creates spatial and nonspatial anndata object from results_df
    """
    params_name = f"scaled_gaussian_pc{pca_dims[0]:2d}_nc{lambda_list[0]:0.2f}_r{resolutions[0]:0.2f}"
    print(params_name)
    adata_spatial = results_df.loc[params_name, "adata"]
    label_name = f"labels_{params_name}"
    adata_spatial.obs['cell type'] = adata_spatial.obs[label_name].map(
        cluster2annotation_spatial
    ).astype('category')

    params_name = f"nonspatial_pc{pca_dims[0]:2d}_nc0.00_r{resolutions[0]:0.2f}"
    print(params_name)
    adata_nonspatial = results_df.loc[params_name, "adata"]
    label_name = f"labels_{params_name}"
    adata_nonspatial.obs['cell type'] = adata_nonspatial.obs[label_name].map(
        cluster2annotation_nonspatial
    ).astype('category')

    return adata_spatial, adata_nonspatial

def calculate_ari(adata, manual:str, predicted:str):
    return adjusted_rand_score(
        adata.obs[manual].cat.codes, 
        adata.obs[predicted].cat.codes
    )


import time, gc
import scanpy as sc
import anndata
from typing import Tuple
import numpy as np
from copy import deepcopy as dcp

sc.settings.verbosity = 1 # Hide information verbose

def print_max_min(adata: anndata.AnnData) -> None:
    '''Displays the max and min of anndata dataset'''
    print("Displaying max and min of Dataset")
    print(f"Max: {adata.X.max()}, Min: {adata.X.min()}\n")


def normalize_total(adata: anndata.AnnData) -> anndata.AnnData:
    '''Normalizes the dataset inplace'''
    
    print("--- Max-Min before normalization -----")
    print_max_min(adata)

    sc.pp.normalize_total(adata, inplace = True)
    print("--- Max-Min after normalization -----")
    print_max_min(adata)
    return adata


def filter_cells(adata: anndata.AnnData,
                 min_count: int,
                 max_count: int,
                 MT_filter: int,
                 gene_filter: int) -> anndata.AnnData:
    
    
    '''
    Inital filter of cells by defining thresholds for min_count, max_count.
    Cells with MT and gene counts below the `MT_filter` and `gene_filter` wil be filtered out
    '''
    
    print(f"Cells before filtering: {adata.n_obs}\n")
    sc.pp.filter_cells(adata, min_counts=min_count)
    sc.pp.filter_cells(adata, max_counts=max_count)
    print(f"Cells after count filter: {adata.n_obs}\n")

    adata = adata[adata.obs["pct_counts_mt"] < MT_filter]
    print(f"cells after MT filter: {adata.n_obs}\n")

    sc.pp.filter_genes(adata, min_cells=gene_filter)
    print(f"Genes after minimum cells per gene filter: {adata.n_vars}\n")
    return adata


def feature_selection(
        adata: anndata.AnnData,
        sample: str,
        coord_keys: Tuple[str],
        hvgs: int = 2000,
        svgs: int = 0,
        load_preprocesssed: bool = True,
        path_to_hvgs: str = None,
        save_genes: bool = False,
        show_figs: bool = False,
        svg_knn: int = 36) -> anndata.AnnData:


        '''
        Parameters:
            adata (anndata): AnnData object containing cell-by-gene

            coord_keys: a tuple of strings to access the spatial coordinates of the anndata object

            hvgs (int): Number of highly variable genes to select

            svgs (int): Number of spatially variable genes to select


        '''
        try:
            from scgft.tl import get_svgs, get_hvgs
        except ModuleNotFoundError:
            print("Error scgft module is not in directory")
            print('Try settings get `svgs = 0`')
            return adata
        
        start_time = time.perf_counter()
    
        if svgs and hvgs:
            print(f'Before filtering: {adata}')
            hvg_genes = get_hvgs(dcp(adata),
                        top_hvgs = hvgs,
                        flavor='seurat')
                        
            svg_genes = get_svgs(dcp(adata),
                        top_svgs = svgs,
                        num_neighbours = svg_knn,
                        spatial_key = coord_keys[2],
                        plot_svg=show_figs)
                        
            adata = adata[:, adata.var['genename'].isin(hvg_genes) | adata.var['genename'].isin(svg_genes)]
            print(f"Number of filtered dataset in SVGs {adata.var['genename'].isin(svg_genes).value_counts()}")
            print(f"Number of filtered dataset in HVGs {adata.var['genename'].isin(hvg_genes).value_counts()}")
            
        elif svgs:
            try:
                from scgft.tl import get_svgs
            except ModuleNotFoundError:
                print("Error scgft module is not in directory to filter by SVGs")
                print('Try settings get `svgs = 0`')
                return adata
            
            svg_genes = get_svgs(dcp(adata),
                                top_svgs = svgs,
                                num_neighbours = svg_knn,
                                spatial_key=coord_keys[2],
                                plot_svg=show_figs)
            
            print(len(svg_genes))
            adata = adata[:, adata.var['genename'].isin(svg_genes)]
            print(adata.var['genename'].isin(svg_genes))

        ### Filter HVGs by the top 1000
        elif hvgs:
            if load_preprocesssed:
                print(f'reading hvg-genes from {path_to_hvgs}')
                hvg_genes = np.loadtxt(path_to_hvgs, delimiter=",", dtype=str)

                adata = adata[:, adata.var['gene_ids'].isin(hvg_genes)]

            else:
                print('Using Scanpy\'s get HVG')
                hvg_genes = get_hvgs(dcp(adata),
                            top_hvgs = hvgs,
                            flavor ='seurat_v3')
                        
                adata = adata[:, adata.var['genename'].isin(hvg_genes)]
        
        print(f"Ran filter in {round(time.perf_counter()-start_time,2)} s")
        print(f'No of SVGs: {svgs}, No of HVGs: {hvgs}')

        if save_genes:
            # Saves list of genes as csv file
            gene_ids =  adata.var['gene_ids'].to_numpy()
            title = f'Filtered_genes_for_sample_{sample}.csv'
            np.savetxt(title, 
                gene_ids,
                delimiter =", ", 
                fmt ='% s')

        gc.collect()

        return adata


def filter_hvg(adata: anndata.AnnData,
               n_top_genes: int,
               flavor: str = "seurat"):
    
    '''Creates a copy of the original annadata object
    Applies log-transformation, 
    then filter by highly-varaiable genes

    Input: 
        anndata: The anndata object 
        n_top_genes: The top highly variable genes to be filtered by 
        flavor = "seurat" or "seurat_v3: 
            if flavor is seurat, a log-transform is applied
            otherwise if flavor is seurat_v3, we do not apply the log-transform
    
    Returns the transformed and filtered dataset
    '''
    ### TO BE EDITED
    if flavor.__eq__("seurat"):
        # If we want to apply log-transform
        adata_log1p = adata.copy()
        sc.pp.log1p(adata_log1p)
        
        print("--- Normalized and log-transformed data -----")
        print_max_min(adata_log1p)

        sc.pp.highly_variable_genes(adata_log1p, 
                                flavor=flavor, 
                                n_top_genes=n_top_genes)
        
        hvg_filter = adata_log1p.var["highly_variable"]
    
    elif flavor.__eq__("seurat_v3"):
        # using the raw count data without log-transform
        print("--- Normalized data -----")
        sc.pp.highly_variable_genes(adata, 
                                flavor=flavor, 
                                n_top_genes=n_top_genes)
        
        hvg_filter = adata.var["highly_variable"]
    
    else:
        print(f"Flavour: {flavor} is not recognized, please use either \'seurat\' or \'seurat_v3\'")
        return

    # save a copy of the unfiltered
    adata_allgenes = adata.copy()
    adata = adata[:, hvg_filter]

    print(f"Displaying dataset after filtering by HVG")
    print(adata)
    return adata, adata_allgenes


'''
This script contains all the utility objects (e.g., dictionaries for padding clusters etc)
for the Slideseq dataset.

NOTE that for data analysis we recommend users to define their own references.
'''

# labels for clusters in dropviz scRNAseq dataset (see dropviz.org)
dropviz_dict = {
    "GranularNeuron_Gabra6": 1,
    "PurkinjeNeuron_Pcp2": 2,
    "Interneurons_Pvalb": 3,
    "Interneurons_and_Other_Nnat": 4,
    "Microglia_Macrophage_C1qb": 5,
    "Oligodendrocyte_Polydendrocyte_Tfr_Tnr": 6,
    "BergmannGlia_Gpr37l1": 7,
    "Astrocyte_Gja1": 8,
    "Choroid_Plexus_Ttr": 9,
    "Endothelial_Flt1": 10,
    "Fibroblast-Like_Dcn": 11,
}

# Slide dropviz reference DE genes
markergenes_dict = {
    "GranularNeuron_Gabra6": [
        "Sh3bgr", "Fat2", "Kcnj12", "Bcl2l15", "Reln", "Uncx", "Plk5", "Tmem163", "Grin2a", "Rims3", "Lrrtm3", "Mctp1",
        "Grin2c", "Ipcef1", "Nr4a2", "Sptb"
    ],
    "PurkinjeNeuron_Pcp2": [
        "Arhgef33", "Ppp1r17", "Nrk", "Clmn", "Slc1a6", "Atp2a3", "Nexn", "Cck", "Camk2a", "Ppp1r16b", "Sycp1", "Itpka",
        "RP23-372N19.2", "Trabd2b", "Gm5083", "Cabp7", "Gm14033", "Cacna1g", "Prkcg", "B3gnt5", "Scn4b", "Cep76",
        "Pde5a", "Shank2", "Kcna6", "Tmem255a", "Prmt8", "Fam174b", "Eps8l2", "Stk17b", "Cmya5", "Plxdc1", "Col18a1",
        "Hes3", "Cntnap5b", "Krt25", "Scd3", "Trpc3", "Atrnl1", "Iltifb", "Kcng4", "Tuba8", "A330050F15Rik", "Shisa6",
        "Bcl11a", "Dagla", "Htr1b", "Nell1", "Ppp4r4", "Cacnb2", "Ebf2", "Zfp385c", "Clip4", "Srgap1", "Arhgap32",
        "Plekhd1", "Grik1", "Epha5", "Kcnip1", "Slc35f1", "Car7", "Ptprm", "Large", "Pcsk6", "Robo2", "Adgrl2", "Lhx5",
        "Ptchd4", "Casq2"
    ],
    "Interneurons_Pvalb": [
        "Kit", "Sorcs3", "Gria3", "Tfap2b", "Slc24a3", "B3galt2", "Prkcd", "Whrn", "Ntn1", "Grik3", "Lypd6", "Ar",
        "Socs2", "Penk", "Nrsn2", "Parm1", "Sla", "Grid1", "N28178", "Nav1", "Plch1", "Esrrg", "Frmpd4", "Hpca",
        "Asic2", "Trim67", "Pak3", "Nxph1", "Vstm2l", "Grin2b", "Nos1ap", "Slc32a1", "Adarb2", "Sema3e", "Skor1",
        "Thsd7a", "Lamb1", "Mar-11", "Neurod6", "Cacna2d3", "Hunk", "Megf10", "Plxnc1", "Ret", "Tmem132e", "Cntn4",
        "A830039N20Rik", "Kcnab3", "Steap2", "Ajap1", "Chst1", "Dusp10", "Mgat5", "Epha8", "Hcn1", "Kcng4", "Psd2",
        "Chd5", "Cttnbp2", "Fam84a", "Grin3a", "Phactr2", "Shisa6", "Btbd11", "Gdpd5", "Prdm8", "Shank2", "Pam",
        "Arhgap32", "Csrnp3", "Sipa1l3", "AI504432", "Galnt18", "Lhfpl3", "Prkca", "Xkr4", "2310067B10Rik"
    ],
    "Interneurons_and_Other_Nnat": [
        "Sln", "Sst", "Nrgn", "Slc17a6", "Doc2g", "Dclk3", "Grm2", "Rgs6", "Vgf", "Grm5", "Tpbg", "Lypd1", "Adcy2",
        "Tcf7l2", "Slc6a5", "Fam19a2", "Neurod6", "Samd3", "Zcchc12", "Pax2", "Basp1", "Rbms3", "Edil3", "Rab3b",
        "Nxph4", "Rasgrp1"
    ],
    "Microglia_Macrophage_C1qb": [
        'C1qb', 'Ctss', 'C1qc', 'C1qa', 'Cx3cr1', 'Csf1r', 'Tyrobp', 'Laptm5', 'Siglech', 'Fcer1g',
        'Trem2', 'Lyz2', 'Fcrls', 'P2ry12', 'Aif1', 'Mrc1', 'Ly86', 'Stab1', 'Selplg', 'Cd74', 'Gpr34',
        'Pf4', 'Fcgr3', 'Rnase4', 'Olfml3', 'Cd68', 'Unc93b1', 'Ctsc', 'Mpeg1', 'Rgs10', 'H2-Ab1',
        'F13a1', 'H2-Eb1',
    ],
    "Oligodendrocyte_Polydendrocyte_Tfr_Tnr": [
        "Mag", "Cldn11", "Mal", "Mog", "Ermn", "Olig1", "Ugt8a", "Opalin", "Gjb1", "Pllp", "Tmem88b", "Gpr37",
        "Gm21984", "Serpinb1a", "Tspan2", "Klk6", "Aspa", "Efnb3", "Sox2ot", "Ppp1r14a", "Gsn", "Pex5l", "Anln", "Il33",
        "Litaf", "Edil3", "Hapln2", "Tnfaip6", "Pdlim2", "Sox10", "Car14", "Fa2h", "Gjc2", "Nkx6-2", "Tmem125",
        "2810468N07Rik", "Arsg", "Prr5l", "Trim59", "Olig2", "Tmem63a", "Adamts4", "Tmeff1", "Grb14", "Lpar1",
        "Carhsp1", "Evi2a", "Gstm7", "Pls1", "Rhog", "Gltp", "Psat1", "Galnt6", "Enpp6", "Cd82", "Cers2", "Elovl7",
        "Pcolce2", "Plcl1", "Plekhh1", "Tmem98", "Tmcc3", "Adssl1", "Il1rap", "Itgb4", "Bcas1", "Frmd8",
        "RP23-240E15.3", "Rtkn", "Tmbim1", "Gal3st1", "Insc", "Prr18", "Jam3", "Cpm", "Myrf", "Erbb2ip", "Fgfr2",
        "Enpp4", "Kctd4", "Sh3gl3", "Neat1", "Plin3", "Elovl1", "Ppfibp2"
    ],
    "BergmannGlia_Gpr37l1": [
        "Gdf10", "Id4", "Tlcd1", "Cacng5", "Npy", "Lcat", "Btbd17", "Lgi4", "Kcnj16", "Cyp2j9", "Slc14a1", "Gas1",
        "Cml1", "Luzp2", "Itih3", "Tubb2b", "Hopx", "Cldn10", "Elovl2", "Tnc", "Gpc6", "Mybpc1", "Cthrc1", "Dao", "A2m",
        "Ddah1", "Col9a3", "Sox2", "Cdc42ep4", "Gjc3", "Cpne2", "Fzd1", "Prex1", "Ctxn3", "Gpld1", "Pax3", "Slc7a10",
        "Abi3bp", "Lfng", "Rab34", "Slc25a18", "Cd70", "Paqr6", "F3", "Mgst1", "Cdc42ep1", "Cxcl14", "Msx2", "Shisa9",
        "Omg", "Aldh1l1", "Dpy19l3", "Proca1", "Npas3", "1700084C01Rik", "3110082J24Rik", "Slc12a4", "Slco4a1",
        "Tspan15", "Dbx2", "Fxyd7", "Gnb4", "Pdlim4", "Rftn2", "Cd302", "Prex2", "Smpdl3a", "Bok", "E130114P18Rik",
        "Fam20a", "Gucy1a3", "Mertk", "S100a6", "Sox1", "Syt10", "Tmem198b", "Pmp22", "Scrg1", "0610040J01Rik",
        "AI464131", "Gabrg1", "Hk2", "Lrrc2", "Plekho2", "Rasl11a", "Trib2", "Btd", "Cables1", "Irx5", "Jam2", "Lgr6",
        "Mob3b", "Slc39a12", "Stk32a", "Tom1l1", "Abhd3", "Fgfr3", "Pou3f3", "Etnppl", "Marcksl1", "Gabra4", "Slc27a1",
        "Efnb2", "2310022B05Rik", "Emp2", "Gng12", "Itgb8", "Laptm4b", "Tmem176b", "Adamts1", "Phgdh", "Rarres2",
        "Srebf1", "Cib1", "Pcdh10", "Rdh5", "Tst", "Adgra1", "Efhd1", "Pigs", "Rorb", "Cyp2d22", "Fscn1", "Gng5",
        "Hdac8", "Klhl13", "Lbh", "Mboat2", "Nfe2l2", "Pbxip1", "Sash1", "Acad8", "Eva1a", "Fgfr2", "Hist1h1c", "Naaa",
        "Rgma", "Soat1", "Sod3", "Ctso", "Dusp6", "Emid1", "Klf3", "Notch1", "Pantr1", "Pcdh17", "Rnaset2a", "Slc13a3",
        "Smox", "Srgap1", "St3gal6", "Abhd4", "Ccdc24", "Cpq", "Nebl", "Pou3f2", "Rab31", "Rhoc", "Rhoj", "Sox5",
        "Zfyve21"
    ],
    "Astrocyte_Gja1": [
        "Hhatl", "Cd38", "Cd44", "Cyp26b1", "Efemp1", "Slc6a11", "Gfap", "Epha4", "Adgrg1", "Eva1a"
    ],
    "Choroid_Plexus_Ttr": [
        'Ccdc153', '1500015O10Rik', 'Gm973', 'Rsph1', 'Mia', 'Tmem212', 'Enkur', 'Dynlrb2',
        'Igfbp2', 'Fam183b', 'Ccdc146', 'Prlr', 'Sostdc1', 'Foxj1', 'Cfap126', 'Sntn', 'Kl',
        'F5', 'Kcnj13', 'Ak7', 'Calml4', 'Wdr66', 'Iqca', '1700016K19Rik', 'Folr1', 'Fhad1',
        'Clic6', 'Rbp1', 'Lrriq1', 'Meig1', 'Cfap45', 'Efcab10', 'Rsph4a', 'Mns1', 'Ezr',
        'Efcab1', 'Kcne2', 'Ccnd2', 'Col8a1', 'Tmem72', 'Car12', 'Cdkn1c', 'Prr32',
    ],
    "Endothelial_Flt1": [
        'Ly6c1', 'Ly6a', 'Cldn5', 'Slco1c1', 'Itm2a', 'Slco1a4', 'Flt1', 'Adgrf5', 'Abcb1a', 'Klf2',
        'Cxcl12', 'Srgn', 'Ctla2a', 'Sdpr', 'Ptprb', 'Palmd', 'Adgrl4', 'Ramp2', 'Wfdc1', 'Emcn',
        'Cd34', 'Egfl7', 'Ifitm3', 'Fxyd5', 'Pecam1', '9430020K01Rik', 'Abcg2', 'Nostrin',
        'Ltbp4', 'Kdr', 'Sema3c', 'Pglyrp1', 'Lef1', 'Lmo2', 'Nes', 'Id1', 'Eng', 'Tm4sf1',
        'Anxa3', 'Tek', 'Acvrl1', 'Ahnak', 'Esam', 'Cav1', 'Lsr', 'C130074G19Rik', 'Fn1',
        'Klf4', 'Uaca', 'Slc22a8', 'Erg', 'BC028528', 'Foxq1', 'Clic5', 'Tmem204', 'Tinagl1',
        'Ets1', 'Podxl', 'Cdh5', 'Ccdc141', 'Ctsh', 'Slc40a1', 'Nrp1', 'Grap', 'Rasgrp3',
        'Slfn5', 'Cd93', 'Cmtm8', 'Kank3', 'Mfsd2a', 'Fli1', 'Epas1', 'Hspb1', 'Gkn3', 'Anxa2',
        'Gimap6', 'Csrp2', 'Isg15', 'Icam2', 'Tmem252', 'Ptrf', 'Gbp7', 'Il10rb', 'Psmb8',
        'Slc38a5', 'Edn3', 'Slc39a8', 'Slc7a5', 'Sox17', 'Arpc1b', 'Degs2', 'Grasp', 'Lgals9',
        'Scarb1', 'Thbd', 'Apold1', 'Trim47', 'AW112010', 'Bst2', 'Fcgrt', 'Sp100', 'Tshz2',
        'AU021092', 'Arhgap29', 'Grrp1', 'Itga6', 'Pdgfb', 'Ifitm2', 'Bcam', 'Gbp3', 'Itga1',
        'Lims2', 'Prom1', 'Slc16a2', 'Tagln2', 'Ddc', 'Edn1', 'Fzd6', 'Rsad2', 'Tie1', 'Bvht',
        'Hmgcs2', 'Net1', 'Pde2a', 'Wwtr1', 'Alpl', 'Gm694', 'H2-T23', 'She', 'Slc16a4', 'Tbx3',
        'Tgm2', 'Rgs12', 'St3gal6', '8430408G22Rik', 'Cyyr1', 'Gata2', 'Ifih1', 'Mmrn2', 'Ndnf',
        'Paqr5', 'Serpinb9', 'Col4a2', 'Ceacam1', 'Cp', 'Elk3', 'Elovl7', 'Foxf2', 'Gimap1',
        'Ppfibp1', 'Ly6c2', 'Robo4', 'Xaf1', 'Afap1l1', 'Col4a1', 'Rasip1', 'Slc7a1', 'Clec14a',
        'Fnbp1l', 'Hspa12b', 'St8sia4', 'Tdrp', 'Tgfbr2', 'Tmem140', 'Fas', 'Gm9946', 'Hmcn1',
        'Ifit1', 'Iigp1', 'Mecom', 'Rtp4', 'Tjp1', 'Ushbp1', 'Abhd2', 'Ocln', 'Atp10a', 'Capg',
        'Eogt', 'F11r', 'Ifi35', 'Lcn2', 'Tspo', 'Unc45b', 'Abcc4', 'Kif26a', 'Pir', 'Slco2b1',
        'St6galnac2', 'Cdkn1c', 'Adcy4', 'Ecscr', 'Filip1', 'Gbp2', 'Heg1', 'Serpinb6b', 'Synm',
        'Thsd1', 'Xdh', 'Cav2', 'Isyna1', 'Rhoc', 'Cnn2', 'Klhl5', 'Prss23', 'Vwf', 'Hes1',
        'Lima1', 'Polk', 'Tmem123', 'Vat1', 'B4galt4', 'Lipa', 'Tns1', 'Txnip', 'Irf9',
        'Swap70', 'Cfh', 'Rassf3',
    ],
    "Fibroblast-Like_Dcn": [
        'Vtn', 'Rgs5', 'Acta2', 'Igf2', 'Myl9', 'Crip1', 'Myh11', 'Dcn', 'Tagln', 'Tpm2', 'Higd1b', 'Lmod1',
        'Mylk', 'Rgs4', 'Filip1l', 'Aebp1', 'Mgp', 'Kcnj8', 'Art3', 'Casq2', 'Ndufa4l2', 'Slc6a20a', 'Aspn',
        'Cox4i2', 'Igfbp2', 'Nupr1', 'Col3a1', 'Pdgfrb', 'Cspg4', 'P2ry14', 'Sod3', 'Ifitm1', 'Ptrf', 'Slc7a11',
        'Sncg', 'Bgn', 'Pcolce', 'Mustn1', 'Slc6a13', 'Gm13861', 'Rbpms', 'Myo1b', 'Abcc9', 'Ace2', 'Cfh', 'Cp',
        'Gja4', 'Ogn', 'Perp', 'S1pr3', 'Bcam', 'Spp1', 'Gpx8', 'Col1a1', 'S100a11', 'Itga1', 'Ifitm2', 'Crispld2',
        'Des', 'Atp13a5', 'Rbp4', 'Ano1', 'Arhgdib', 'Olfr558', 'Pln', 'Snhg18', 'Nr2f2',
        'Gpc3', 'Rcn3', 'Cnn2', 'Emp3', 'Nexn', 'Adap2', 'Cd248', 'Col4a1', 'Foxc1', 'Slc22a6', 'Tagln2',
        'Hspb1', 'Enpep', 'Fmod', 'Serping1', 'Cdkn1c', 'Heyl', 'Efemp1', 'Gjb2', 'Colec12', 'Phldb2',
        'Wls', 'Arhgap29', 'Col4a2', 'Il34', 'Nid1', 'Slc13a4', 'Aldh1a2', 'Lum', 'Pde1a', 'Tbx18', 'Tinagl1',
        'Flna', 'Serpinf1', 'Rbp1', 'Lamb2', 'Map3k7cl', 'C1qtnf2', 'Crabp2', 'Anxa2', 'Isyna1',
        'Fbln1', 'Islr', 'Slc13a3',
    ],
}



"""
Functions for plotting

Nigel 3 dec 2020

Modified 16 May 2023
"""

from typing import Union, Tuple, Collection

import os
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.collections import LineCollection
from matplotlib.patches import Rectangle
import seaborn as sns

from scipy.sparse import csr_matrix

import warnings

from banksy_utils.time_utils import timer
from banksy.labels import Label


@timer
def plot_edge_histogram(graph: csr_matrix,
                        ax: mpl.axes.Axes,
                        title: str = "edge weights",
                        bins: int = 100):
    """
    plot a histogram of the edge-weights a graph
    """
    counts, bins, patches = ax.hist(graph.data, bins=bins)

    median_dist = np.median(graph.data)
    mode_dist = bins[np.argmax(counts)]
    ax.axvline(median_dist, color="r", alpha=0.8)
    ax.axvline(mode_dist, color="g", alpha=0.8)
    ax.set_title("Histogram of " + title)

    print(f"\nEdge weights ({title}): "
          f"median = {median_dist}, mode = {mode_dist}\n")

    return median_dist, mode_dist


@timer
def plot_2d_embeddings(embedding: np.ndarray,
                       labels: Union[np.ndarray, list],
                       method_str: str = "2D projection",
                       space_str: str = "of some space",
                       xlabel: str = "Dimension 1",
                       ylabel: str = "Dimension 2",
                       ax: mpl.axes.Axes = None,
                       plot_cmap: bool = True,
                       cmap_name: str = "Spectral",
                       figsize: tuple = (8, 8),
                       title_fontsize: int = 12,
                       **kwargs,
                       ) -> None:
    """
    plot embeddings in 2D, with coloured labels for each point

    :param embedding: the embedding matrix (only 1st 2 columns are used)
    :param labels: integer labels for each point
    """
    num_labels = len(np.unique(labels))

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        ax = ax
        fig = ax.get_figure()

    scatterplot = ax.scatter(
        embedding[:, 0], embedding[:, 1],
        c=labels,
        cmap=cmap_name,
        s=0.2, alpha=0.5,
        **kwargs,
    )

    # ax.set_aspect("equal", "datalim")
    # ax.set_aspect("datalim")
    ax.set_title(f"{method_str} {space_str}",
                 fontsize=title_fontsize)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    print(f"number of labels: {num_labels}")

    if plot_cmap:
        fig.colorbar(
            scatterplot,
            boundaries=np.arange(num_labels + 1) - 0.5
        ).set_ticks(np.arange(num_labels))


@timer
def plot_graph_weights(locations,
                       graph,
                       theta_graph=None,  # azimuthal angles
                       max_weight=1,
                       markersize=1,
                       figsize=(8, 8),
                       title: str = None,
                       flip_yaxis: bool = False,
                       ax=None,
                       ) -> None:
    """
    Visualize weights in a spatial graph, 
    heavier weights represented by thicker lines
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    edges, weights, theta = [], [], []

    if theta_graph is not None:
        assert isinstance(theta_graph, csr_matrix)
        assert theta_graph.shape[0] == graph.shape[0]

    for start_node_idx in range(graph.shape[0]):

        ptr_start = graph.indptr[start_node_idx]
        ptr_end = graph.indptr[start_node_idx + 1]

        for ptr in range(ptr_start, ptr_end):
            end_node_idx = graph.indices[ptr]

            # append xs and ys as columns of a numpy array
            edges.append(locations[[start_node_idx, end_node_idx], :])
            weights.append(graph.data[ptr])
            if theta_graph is not None:
                theta.append(theta_graph.data[ptr])

    print(f"Maximum weight: {np.amax(np.array(weights))}\n")
    weights /= np.amax(np.array(weights))

    if theta_graph is not None:
        norm = mpl.colors.Normalize(vmin=min(theta), vmax=max(theta), clip=True)
        mapper = mpl.cm.ScalarMappable(norm=norm, cmap="bwr")
        c = [mapper.to_rgba(t) for t in theta]
    else:
        c = "C0"

    line_segments = LineCollection(
        edges, linewidths=weights * max_weight, linestyle='solid', colors=c, alpha=0.7)
    ax.add_collection(line_segments)

    ax.scatter(locations[:, 0], locations[:, 1], s=markersize, c="gray", alpha=.6, )

    if flip_yaxis:  # 0 on top, typical of image data
        ax.set_ylim(ax.get_ylim()[::-1])

    ax.set_aspect('equal', 'datalim')

    if title is not None:
        ax.set_title(title)


def plot_continuous(x, y, c, ax,
                    spot_size: int = 0.3,
                    vmax: float = None,
                    cmap: str = "Blues",
                    title: str = None,
                    plot_cbar: bool = True,
                    ) -> None:
    """
    Plot continous-valued location plot (e.g. markergene, RCTD weights)
    """
    scatterplot = ax.scatter(x, y, c=c,
                             s=spot_size, alpha=1.0,
                             cmap=cmap,
                             vmax=vmax, vmin=0,
                             )

    ax.set_aspect('equal', 'datalim')
    if title is not None:
        ax.set_title(title, fontsize=8, fontweight="bold", )
    # ax.set_ylim(ax_locs.get_ylim()[::-1])
    ax.axis("off")

    if plot_cbar:
        cax = ax.inset_axes([0.8, 0.0, 0.16, 0.04])
        # cax.axis("off")
        ax.figure.colorbar(scatterplot, cax=cax, orientation='horizontal')


def plot_cluster_subset(cluster_subset: list,
                        colours_dict: dict,
                        adata,
                        ax,
                        title: str = None,
                        umap_name: str = "umap",
                        spot_size: float = 0.4,
                        plot_umap: bool = False,
                        ):
    """
    Plot a subset of clusters using specified colours from colours_dict
    """

    def plot_embedding(embedding, ax):

        for cluster in adata.obs["cell type"].cat.categories:

            if cluster not in cluster_subset:
                c = "whitesmoke"
                locations = adata.obsm[embedding][adata.obs["cell type"] == cluster, :]
                scatterplot = ax.scatter(locations[:, 0], locations[:, 1], c=c,
                                         s=spot_size, alpha=1.0)

        for cluster in adata.obs["cell type"].cat.categories:

            if cluster in cluster_subset:
                c = colours_dict[cluster]
                locations = adata.obsm[embedding][adata.obs["cell type"] == cluster, :]
                scatterplot = ax.scatter(locations[:, 0], locations[:, 1], c=c,
                                         s=spot_size, alpha=1.0)

    # plot locations
    plot_embedding("coord_xy", ax)
    ax.set_aspect('equal', 'datalim')
    if title is not None:
        ax.set_title(title, fontsize=20, fontweight="bold", )
    # ax.set_ylim(ax_locs.get_ylim()[::-1])
    ax.axis("off")

    # plot  UMAP as inset
    if plot_umap:
        axins = ax.inset_axes([0.9, 0.0, 0.25, 0.25])
        plot_embedding(umap_name, axins)
        axins.set_aspect('equal', 'datalim')
        # axins.axis("off")
        axins.set_ylabel("UMAP2", fontsize=8, fontweight="bold", )
        axins.set_xlabel("UMAP1", fontsize=8, fontweight="bold", )
        axins.spines['right'].set_visible(False)
        axins.spines['top'].set_visible(False)
        axins.xaxis.set_ticks([])
        axins.yaxis.set_ticks([])


@timer
def plot_genes(genes: list,
               df: pd.DataFrame,
               results_df: pd.DataFrame = None,  # only used for plotting spatialDE length-scales
               x_colname: str = "X",
               y_colname: str = "  Y",
               plots_per_row: int = 4,
               colormap: str = "Blues",
               vmin: float = None,
               vmax: float = None,
               take_log: bool = False,
               main_title: str = None,
               verbose: bool = True,
               ):
    """
    Plot the count distributions of a list of genes
    """

    num_genes = len(genes)
    num_rows = len(genes) // plots_per_row + 1

    print(
        f"Number of genes: {num_genes}\n"
        f"Plots per row: {plots_per_row}\n"
        f"Number of rows: {num_rows}"
    )

    fig = plt.figure(figsize=(6 * plots_per_row, 4 * num_rows))
    fig.subplots_adjust(hspace=0.4, wspace=0.4)

    df_columns = df.columns

    sns.set_style("white")

    for gene_index, gene in enumerate(genes):

        if gene not in df_columns:
            warnings.warn(f"Gene {gene} not found in dataframe!")
            continue

        ax = fig.add_subplot(num_rows, plots_per_row, gene_index + 1)
        x = df[x_colname].values
        y = df[y_colname].values
        counts = df[gene].values

        if take_log:
            counts = np.log10(counts + 1)

        ax.scatter(
            x, y, s=10, c=counts,
            cmap=colormap,
            vmin=vmin, vmax=vmax,
            alpha=0.9,
        )

        df.plot.scatter(
            x=x_colname, y=y_colname, s=10, c=gene,
            colormap=colormap,
            vmin=vmin, vmax=vmax, alpha=0.9,
            ax=ax
        )

        if results_df is not None:

            # Get the set of p-values (1 for each length)
            # -------------------------------------------

            gene_results = results_df.loc[results_df["g"] == gene, ["l", "pval"]]
            gene_results.sort_values(by="l", inplace=True)

            pvals = -1 * np.log10(gene_results["pval"].values)
            # pvals = gene_results["pval"].values
            # normalize to the highest p-value
            pvals /= np.max(pvals)
            num_boxes = len(pvals)

            y_lower, y_upper = ax.get_ylim()
            x_lower, x_upper = ax.get_xlim()
            y_span = np.abs(y_upper - y_lower)
            x_span = np.abs(x_upper - x_lower)
            y_step = y_span / num_boxes
            x_step = x_span / 20

            if verbose:
                print(f"{gene_results}\n"
                      f"P-values:\n{pvals}"
                      f"\n y lower limit : {y_lower}, y upper limit : {y_upper}"
                      f"\n x lower limit : {x_lower}, x upper limit : {x_upper}")

            cmap = plt.cm.get_cmap('viridis')

            for box_num in range(num_boxes):
                rect = Rectangle((x_upper, y_lower + y_step * box_num), x_step, y_step,
                                 linewidth=1, facecolor=cmap(pvals[box_num]), edgecolor="gray",
                                 clip_on=False, )
                ax.add_patch(rect)

        # Title
        # -----

        if "blank" in gene or "Blank" in gene:
            title_colour = "r"
        else:
            title_colour = "b"

        ax.set_title(gene,
                     fontdict={'fontsize': 20,
                               'fontweight': 'bold',
                               'color': title_colour, }
                     )
        # plt.gca().invert_yaxis()
        ax.axis('equal')

    if main_title is not None:
        fig.suptitle(main_title, fontsize=25,
                     # fontproperties={'fontsize': 25,
                     #                 'fontweight': 'bold',
                     #                 }
                     )

    plt.show()


def plot_labels_seperately(label: Label,
                           locations: np.ndarray,
                           embeddings: np.ndarray = None,
                           embedding_str: str = "UMAP",
                           plots_per_row: int = 3,
                           cmap_name: str = "Spectral",
                           colour_list: list = None,
                           max_id: int = None,
                           default_colour: str = "tab:red",
                           background_colour: str = "whitesmoke",
                           subplot_size: Tuple[float, float] = (5, 5),
                           spot_size: float = 0.2,
                           alpha: float = 1,
                           flip_axes: bool = False,
                           show_axes: bool = False,
                           save_fig: bool = False,
                           save_path: str = None,
                           verbose: bool = False,
                           ) -> None:
    """
    Plot each label's locations individually in subplots
    (nonlabel locations in gray)
    If embeddings is not None, UMAP or other embedding will be plotted.

    label: Label object
    locations: x/y coordinates
    embeddings: embedding coordinates (2D) e.g. UMAP or t-SNE
    embedding_str: type of embedding
    cmap_name: name of standard colourmap or reference to custom colourmap
    max_id: highest cluster ID if not present in current clusters (e.g. when matching labels)
    default_colour: colour for clusters if no colourmap given
    background_colour: colour for spots not in cluster
    subplot_size: size of individual subplot
    flip_axes: flip y axis
    show_axes: show x and y axis borders
    """

    num_rows = label.num_labels // plots_per_row + 1
    figsize_x = (subplot_size[0] * plots_per_row) * 1.1
    if embeddings is not None:
        figsize_x *= 2
    figsize_y = subplot_size[1] * num_rows

    if verbose:
        print(f"\nGenerating {num_rows} by {plots_per_row} "
              f"grid of plots (figsize = {figsize_y}, {figsize_x})\n"
              f"Plotting labels: {label.ids}, "
              f"max ID: {max(max_id, label.max_id)}\n")

    fig = plt.figure(figsize=(figsize_x, figsize_y), constrained_layout=True)

    if embeddings is not None:
        grid = fig.add_gridspec(ncols=2 * plots_per_row, nrows=num_rows)
    else:
        grid = fig.add_gridspec(ncols=plots_per_row, nrows=num_rows)

    if cmap_name is not None:
        cmap = mpl.cm.get_cmap(cmap_name)

    if max_id is None:
        max_id = label.max_id

    # Plot each label
    # ---------------

    for n, label_id in enumerate(label.ids):

        grid_y = n // plots_per_row
        grid_x = n % plots_per_row

        if verbose:
            print(f"Plotting grid position: {grid_y}, {grid_x}")

        if embeddings is not None:

            ax = fig.add_subplot(grid[grid_y, 2 * grid_x])
            ax_umap = fig.add_subplot(grid[grid_y, 2 * grid_x + 1])

        else:

            ax = fig.add_subplot(grid[grid_y, grid_x])

        onehot = label.get_onehot()
        label_mask = np.squeeze(onehot[n, :].toarray().astype(bool))

        if cmap_name is None:
            if colour_list is not None:
                assert label_id < len(colour_list)
                c = colour_list[label_id]
            else:
                c = default_colour
        else:
            # matplotlib says to add an additional axis (1 x 3 shape)
            # print(f"{label_id + 0.5}/{max_id + 1}={(label_id + 0.5) / (max_id + 1)}")
            c = np.expand_dims(cmap((label_id + 0.5) / (max_id + 1)), axis=0)

        other_spots = ax.scatter(
            locations[~label_mask, 0], locations[~label_mask, 1],
            c=background_colour, s=spot_size, alpha=alpha,
        )

        cluster_spots = ax.scatter(
            locations[label_mask, 0], locations[label_mask, 1],
            c=c, s=spot_size, alpha=alpha,
        )

        if flip_axes:
            ax.set_ylim(ax.get_ylim()[::-1])
        ax.set_aspect('equal', 'datalim')
        ax.set_title(f'{label_id}', fontsize=4 * subplot_size[0], fontweight="bold")

        # Plot UMAP
        # ---------

        if embeddings is not None:
            umap_other = ax_umap.scatter(
                embeddings[~label_mask, 0], embeddings[~label_mask, 1],
                c=background_colour, s=spot_size, alpha=alpha,
            )

            umap_cluster = ax_umap.scatter(
                embeddings[label_mask, 0], embeddings[label_mask, 1],
                c=c, s=spot_size, alpha=alpha,
            )

        # ax_umap.set_aspect('equal', 'datalim')
        ax_umap.set_title(f'{label_id} {embedding_str}',
                          fontsize=4 * subplot_size[0], fontweight="bold")
        ax_umap.set_xlabel(embedding_str + "1")
        ax_umap.set_ylabel(embedding_str + "2")

        if not show_axes:
            ax.axis("off")
            ax_umap.axis("off")

    if save_fig and isinstance(save_path, str):
        fig.savefig(save_path)


def plot_label_subset(labels_to_plot: Collection,
                      label: Label,
                      locations: np.ndarray,
                      embeddings: np.ndarray,
                      cmap_name: str = "Spectral",
                      max_id: int = None,
                      figsize: Tuple[float, float] = (10, 5),
                      verbose: bool = True,
                      save_fig: bool = False,
                      save_path: str = None,
                      ) -> None:
    """
    Plot a subset of labels by colour
    Remaining labels will be shown in gray
    :param labels_to_plot: subset of labels to plot
    :param label: Label object
    :param locations: spatial coordinates
    :param embeddings: UMAP/PCA
    :param max_id: max label id (if None, use max id of Label object)
    """

    cmap = mpl.cm.get_cmap(cmap_name)

    if max_id is None:
        max_id = label.max_id

    labels_to_plot = [label_id for label_id in labels_to_plot
                      if label_id in label.ids]

    if len(labels_to_plot) == 0:
        warnings.warn("labels provided not found in Label set")
        return

    if verbose:
        print(f"plotting labels: {labels_to_plot}")

    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=figsize)

    all_labels_mask = np.zeros((label.num_samples,), dtype=bool)

    # Plot each label by colour
    # -------------------------

    for label_id in labels_to_plot:
        n = list(label.ids).index(label_id)
        print(f"\nPlotting label {label_id} (row {n})")

        label_mask = np.squeeze(label.onehot[n, :].toarray().astype(bool))

        all_labels_mask = all_labels_mask | label_mask

        c = np.expand_dims(cmap(label_id / max_id), axis=0)

        ax[0].scatter(
            locations[label_mask, 0],
            locations[label_mask, 1],
            c=c, s=0.8, alpha=1,
        )

        ax[1].scatter(
            embeddings[label_mask, 0],
            embeddings[label_mask, 1],
            c=c, s=0.8, alpha=1,
        )

    # Plot remaining locations in Gray
    # --------------------------------

    ax[0].scatter(
        locations[~all_labels_mask, 0],
        locations[~all_labels_mask, 1],
        c="whitesmoke", s=0.8, alpha=1,
    )

    ax[1].scatter(
        embeddings[~all_labels_mask, 0],
        embeddings[~all_labels_mask, 1],
        c="whitesmoke", s=0.8, alpha=1,
    )

    if save_fig and isinstance(save_path, str):
        fig.savefig(save_path)


if __name__ == "__main__":
    print("test")



import anndata
import numpy as np
import pandas as pd

from sklearn.metrics import adjusted_rand_score
from sklearn.neighbors import NearestNeighbors
import scipy.sparse as sp

from banksy.labels import Label
from banksy_utils.color_lists import spagcn_color
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import copy
import os, time, gc


def refine_clusters(adata: anndata.AnnData,
                    results_df: pd.DataFrame,
                    coord_keys: tuple,
                    color_list: list = spagcn_color,
                    savefig: bool = False,
                    output_folder: str = "",
                    refine_method: str = "once",
                    refine_iterations: int = 1,
                    annotation_key: str = "manual_annotations",
                    num_neigh: int = 6,
                    verbose: bool = False
                    ) -> pd.DataFrame:
    '''
    Function to refine predicted labels based on nearest neighbours

    Args:
        adata (AnnData): Original anndata object

        results_df (pd.DataFrame): DataFrame object containing the results from BANKSY

    Optional Args: 
        color_list (list); default = spagcn : List in which colors are used to plot the figures. 

        refine_method ("auto" | "once" | "iter_num" ): 
                To refine clusters once only or iteratively refine multiple times
                If "auto" is specified, the refinement procedure completes iteratively 
                    until only 0.5% of the nodes are changed.

                If "iter_num" is specified, specify the 'refine_iterations' parameter

        num_neigh (int) : Number of nearest-neighbours the refinement is conducted over

        savefig (bool): To save figure in folder or not.

        annotation_key (str): The key in whicb the ground truth annotations are accessed under 
                adata.obs[annotation_key],

                If no ground truth is present, then set annotation_key = None

    '''

    start_time = time.perf_counter()

    if annotation_key:
        annotations = adata.obs[annotation_key].cat.codes.tolist()
    else:
        annotations = None

    refined_ari_results, total_entropy_dict = {}, {}

    for params_name in results_df.index:
        adata_temp = results_df.loc[params_name, 'adata']
        raw_labels = results_df.loc[params_name, 'labels']
        num_clusters = results_df.loc[params_name, 'num_labels']
        lambda_p = results_df.loc[params_name, "lambda_param"]

        if annotation_key:
            ari_temp = results_df.loc[params_name, "ari"]

        if isinstance(raw_labels, Label):
            raw_labels = raw_labels.dense

        if coord_keys[2] not in adata_temp.obsm_keys():
            adata_temp.obsm[coord_keys[2]] = pd.concat([adata.obs[coord_keys[0]], adata.obs[coord_keys[1]]],
                                                       axis=1).to_numpy()

        # Refine Clusters
        refined_list = raw_labels[:]

        if refine_method.__eq__("auto"):
            total_entropy, num_iter = 1.0, 1
            # We keep on refining until only 1.0% of the nodes are swapped
            if verbose:
                print(f'Auto Refine')
            while ((total_entropy * 100 > 5) and (num_iter < 20)):
                refined_list, refined_ari, total_entropy = refine_once(adata_temp,
                                                                       refined_list,
                                                                       annotations,
                                                                       coord_keys,
                                                                       num_neigh)
                if verbose:
                    print(f'Refined iteration: {num_iter} | Total Entropy: {round(total_entropy, 2)}')
                num_iter += 1

        elif refine_method.__eq__("once"):
            print(f'Refine only once')
            refined_list, refined_ari, total_entropy = refine_once(adata_temp,
                                                                   refined_list,
                                                                   annotations,
                                                                   coord_keys,
                                                                   num_neigh)
            print(f'Refined once | Total Entropy: {round(total_entropy, 2)}')

        elif refine_method.__eq__("num_iter"):
            # Refine a specifed number of iterations  
            print(f'Refine {refine_iterations} times')
            for i in range(refine_iterations):
                refined_list, refined_ari, total_entropy = refine_once(adata_temp,
                                                                       refined_list,
                                                                       annotations,
                                                                       coord_keys,
                                                                       num_neigh)

                if verbose:
                    print(f'Refined iteration: {i} | Total Entropy: {round(total_entropy, 2)}')

        else:
            print("No valid refine type specified! {\"auto\", \"once\", \"num_iter\"}")
            return results_df

        if annotation_key:
            refined_ari_results[params_name] = refined_ari
            raw_ari = f'\nARI = {round(ari_temp, 2)}'
            refined_ari = f'\nRefined ARI = {round(refined_ari, 2)}'

        total_entropy_dict[params_name] = total_entropy / np.log(num_clusters)

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 8), constrained_layout=True)
        # fig = plt.figure(figsize=main_figsize, constrained_layout=True)
        grid = fig.add_gridspec()
        raw_clusters, refined_clusters = [], []

        for i in raw_labels:
            raw_clusters.append(color_list[i])

        if annotation_key:
            title1 = f'λ = {round(lambda_p, 2)}' + raw_ari
            title2 = f'λ = {round(lambda_p, 2)}' + refined_ari
        else:
            title1 = f'λ = {round(lambda_p, 2)}'
            title2 = f'λ = {round(lambda_p, 2)}'

        subplot_sc(ax1,
                   adata_temp,
                   coord_keys,
                   raw_clusters,
                   num_clusters,
                   title1)

        for i in refined_list:
            refined_clusters.append(color_list[i])

        subplot_sc(ax2,
                   adata_temp,
                   coord_keys,
                   refined_clusters,
                   num_clusters,
                   title2)

        if savefig:
            print(f'{params_name} saved refined plot at {output_folder}')
            save_name = os.path.join(output_folder, str(params_name) + '.png')
            fig.savefig(save_name, format='png', dpi=108)

    if annotation_key:
        results_df["refined_ari"] = refined_ari_results

    results_df["total_Entropy_normalized"] = total_entropy_dict
    gc.collect()
    print(f'Time taken for refinement = {round((time.perf_counter() - start_time) / 60, 2)} min')
    return results_df


def subplot_sc(ax: plt.Axes,
               adata_temp: anndata.AnnData,
               coord_keys: tuple,
               clusters_colormap: list,
               num_clusters: int,
               title: str) -> None:
    ax.scatter(adata_temp.obs[coord_keys[0]],
               adata_temp.obs[coord_keys[1]],
               c=clusters_colormap,
               vmin=0, vmax=num_clusters - 1,
               s=50, alpha=1.0)

    ax.set_aspect('equal', 'datalim')
    ax.set_title(title, fontsize=20, fontweight="bold", )
    ax.axes.invert_yaxis()

    # Turn of ticks and frame
    ax.set(frame_on=False)
    ax.xaxis.set_major_locator(ticker.NullLocator())
    ax.yaxis.set_major_locator(ticker.NullLocator())


def refine_once(adata_temp: anndata.AnnData,
                raw_labels: list,
                truth_labels: list,
                coord_keys: tuple,
                num_neigh: int = 6,
                threshold_percentile: int = None):
    '''Refined clusters with respect to their mean neighbourhood labels,
    if more than 50% of a node's neighbours is of the same label, 
    we swap this node's label with theirs'''

    refined_list = copy.deepcopy(raw_labels)
    nbrs = NearestNeighbors(algorithm='ball_tree').fit(adata_temp.obsm[coord_keys[2]])
    csr_mat = nbrs.kneighbors_graph(n_neighbors=num_neigh, mode='distance')

    if threshold_percentile:
        # if a 'threshold_percentile' is set,
        # cut off edges that are longer than the threshold_percentile
        nnz_inds = csr_mat.nonzero()
        threshold = np.quantile(csr_mat.data, threshold_percentile)
        keep = np.where(csr_mat.data < threshold)[0]
        n_keep = len(keep)
        csr_mat = sp.csr_matrix((np.ones(n_keep), (nnz_inds[0][keep], nnz_inds[1][keep])), shape=csr_mat.shape)

    indptr = csr_mat.indptr
    num_nodes = len(indptr) - 1

    num_nodes_swapped = 0
    total_entropy = 0

    for i in range(num_nodes):
        nbrs = csr_mat[i]
        nbrs_len = len(nbrs.indices)

        if nbrs_len == 0:
            # If node has no neighbours, we skip refinement process
            continue

        current_label = raw_labels[i]  # Node's current label
        nbr_labels = [current_label]

        for n in nbrs.indices:
            # Constructs list of the neighbourhood's labels
            nbr_labels.append(raw_labels[n])

        # Count the number of unqiue labels
        unique, counts = np.unique(nbr_labels, return_counts=True)
        max_counts = np.max(counts)
        res = unique[np.argmax(counts)]  # Majority of neighourhood expression

        # Threshold count for refinement process
        threshold_count = (nbrs_len // 2) + 1
        if (res != current_label) and (max_counts >= threshold_count):
            num_nodes_swapped += 1
            refined_list[i] = res

        # Calculate the (local) neighbourhood's Entropy
        local_entropy = entropy(counts)
        total_entropy += local_entropy

    refined_ari = adjusted_rand_score(refined_list, truth_labels) if truth_labels else -1

    swapped_ratio = num_nodes_swapped / num_nodes
    total_entropy /= num_nodes

    print(f'\nNumber of nodes swapped {num_nodes_swapped} | ratio: {round(swapped_ratio, 3)}')
    print(f'Total Entropy: {round(total_entropy, 2)}')

    return refined_list, refined_ari, total_entropy


def entropy(counts: list) -> float:
    '''Takes a list of labels and outputs the cross Entropy'''
    total_count = np.sum(counts)
    p = counts / total_count
    return np.sum(-p * np.log(p))
